#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/tads3/mkchrtab.cpp,v 1.3 1999/07/11 00:46:58 MJRoberts Exp $";
#endif

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  mkchrtab.cpp - TADS character table generator
Function
  
Notes
  
Modified
  10/17/98 MJRoberts  - creation (from TADS 2 mkchrtab.c)
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "t3std.h"

/*
 *   Read a number.  Returns zero on success, non-zero on error.
 */
static int read_number(unsigned int *result, char **p,
                       char *infile, int linenum, int optional)
{
    unsigned int base;
    unsigned int acc;
    int digcnt;
    
    /* skip any leading spaces */
    while (isspace(**p))
        ++(*p);

    /* 
     *   if the entry is optional, and we've reached a comment or the end
     *   of the line, return failure, but don't print an error (since the
     *   number is optional, it's not an error if it's not present) 
     */
    if (optional &&
        (**p == '\0' || **p == '\n' || **p == '\r' || **p == '#'))
        return 2;

    /*
     *   Check for a character value 
     */
    if (**p == '\'')
    {
        unsigned char c;
        
        /* get the next character */
        ++(*p);

        /* if it's a backslash, escape the next character */
        if (**p == '\\')
        {
            /* skip the backslash */
            ++(*p);

            /* the next character gives the value */
            switch(**p)
            {
            case 'n':
                c = '\n';
                break;

            case 'r':
                c = '\r';
                break;

            case 't':
                c = '\t';
                break;

            case '\\':
                c = '\\';
                break;

            case '\'':
                c = '\'';
                break;

            default:
                printf("%s: line %d: invalid backslash sequence '\\%c'\n",
                       infile, linenum, **p);
                return 4;
            }
        }
        else
        {
            /* use this value as the character code */
            c = (unsigned char)**p;
        }

        /* skip the character, and make sure we have the closing quote */
        ++(*p);
        if (**p != '\'')
        {
            printf("%s: line %d: missing close quote\n", infile, linenum);
            return 5;
        }

        /* skip the close quote */
        ++(*p);

        /* skip trailing spaces */
        while (isspace(**p))
            ++(*p);

        /* return the result */
        *result = (unsigned int)c;
        return 0;
    }

    /* 
     *   determine the base - if there's a leading zero, it's hex or
     *   octal; otherwise, it's decimal 
     */
    if (**p == '0')
    {
        /* skip the leading zero */
        ++(*p);
        
        /* if the next character is 'x', it's hex, otherwise it's octal */
        if (**p == 'x' || **p == 'X')
        {
            /* skip the 'x' */
            ++(*p);

            /* it's hex */
            base = 16;
        }
        else
        {
            /* it's octal */
            base = 8;
        }
    }
    else
    {
        /* no leading zero - it's a decimal number */
        base = 10;
    }

    /* read digits */
    for (acc = 0, digcnt = 0 ;; ++(*p), ++digcnt)
    {
        /* see if we still have a digit */
        if (isdigit(**p) || (base == 16 && isxdigit(**p)))
        {
            unsigned int dig;
            
            /* get this digit's value */
            dig = (unsigned int)(isdigit(**p)
                                 ? **p - '0'
                                 : **p - (isupper(**p) ? 'A' : 'a') + 10);

            /* make sure it's in range for our radix */
            if (dig >= base)
            {
                printf("%s: line %d: invalid digit for radix\n",
                       infile, linenum);
                return 3;
            }

            /* accumulate this digit */
            acc *= base;
            acc += dig;
        }
        else
        {
            /* that's the end of the number */
            break;
        }
    }

    /* if we didn't get any valid digits, it's an error */
    if (digcnt == 0)
    {
        printf("%s: line %d: invalid number\n", infile, linenum);
        return 1;
    }

    /* skip any trailing spaces */
    while (isspace(**p))
        ++(*p);

    /* return the accumulated value */
    *result = acc;

    /* success */
    return 0;
}


/*
 *   Check to see if an identifier matches a given string 
 */
static int id_matches(const char *idp, size_t idlen, const char *str)
{
    return (idlen == strlen(str) && memicmp(idp, str, idlen) == 0);
}

/*
 *   HTML Entity mapping structure 
 */
struct entity_t
{
    const char *ename;
    unsigned int charval;
};

/*
 *   List of HTML TADS entity names and the corresponding Unicode
 *   character codes 
 */
static const struct entity_t entities[] =
{
    { "nbsp", 160 },
    { "iexcl", 161 },
    { "cent", 162 },
    { "pound", 163 },
    { "curren", 164 },
    { "yen", 165 },
    { "brvbar", 166 },
    { "sect", 167 },
    { "uml", 168 },
    { "copy", 169 },
    { "ordf", 170 },
    { "laquo", 171 },
    { "not", 172 },
    { "shy", 173 },
    { "reg", 174 },
    { "macr", 175 },
    { "deg", 176 },
    { "plusmn", 177 },
    { "sup2", 178 },
    { "sup3", 179 },
    { "acute", 180 },
    { "micro", 181 },
    { "para", 182 },
    { "middot", 183 },
    { "cedil", 184 },
    { "sup1", 185 },
    { "ordm", 186 },
    { "raquo", 187 },
    { "frac14", 188 },
    { "frac12", 189 },
    { "frac34", 190 },
    { "iquest", 191 },
    { "times", 215 },
    { "divide", 247 },
    { "AElig", 198 },
    { "Aacute", 193 },
    { "Acirc", 194 },
    { "Agrave", 192 },
    { "Aring", 197 },
    { "Atilde", 195 },
    { "Auml", 196 },
    { "Ccedil", 199 },
    { "ETH", 208 },
    { "Eacute", 201 },
    { "Ecirc", 202 },
    { "Egrave", 200 },
    { "Euml", 203 },
    { "Iacute", 205 },
    { "Icirc", 206 },
    { "Igrave", 204 },
    { "Iuml", 207 },
    { "Ntilde", 209 },
    { "Oacute", 211 },
    { "Ocirc", 212 },
    { "Ograve", 210 },
    { "Oslash", 216 },
    { "Otilde", 213 },
    { "Ouml", 214 },
    { "THORN", 222 },
    { "Uacute", 218 },
    { "Ucirc", 219 },
    { "Ugrave", 217 },
    { "Uuml", 220 },
    { "Yacute", 221 },
    { "aacute", 225 },
    { "acirc", 226 },
    { "aelig", 230 },
    { "agrave", 224 },
    { "aring", 229 },
    { "atilde", 227 },
    { "auml", 228 },
    { "ccedil", 231 },
    { "eacute", 233 },
    { "ecirc", 234 },
    { "egrave", 232 },
    { "eth", 240 },
    { "euml", 235 },
    { "iacute", 237 },
    { "icirc", 238 },
    { "igrave", 236 },
    { "iuml", 239 },
    { "ntilde", 241 },
    { "oacute", 243 },
    { "ocirc", 244 },
    { "ograve", 242 },
    { "oslash", 248 },
    { "otilde", 245 },
    { "ouml", 246 },
    { "szlig", 223 },
    { "thorn", 254 },
    { "uacute", 250 },
    { "ucirc", 251 },
    { "ugrave", 249 },
    { "uuml", 252 },
    { "yacute", 253 },
    { "thorn", 254 },
    { "yuml", 255 },

    /* TADS extensions to the standard characters */
    { "lsq", 8216 },
    { "rsq", 8217 },
    { "ldq", 8220 },
    { "rdq", 8221 },
    { "endash", 8211 },
    { "emdash", 8212 },
    { "trade", 8482 },

    /* HTML 4.0 character extensions */
    { "ndash", 8211 },
    { "mdash", 8212 },
    { "lsquo", 8216 },
    { "rsquo", 8217 },
    { "ldquo", 8220 },
    { "rdquo", 8221 },
    { "sbquo", 8218 },
    { "bdquo", 8222 },
    { "lsaquo", 8249 },
    { "rsaquo", 8250 },
    { "dagger", 8224 },
    { "Dagger", 8225 },
    { "OElig", 338 },
    { "oelig", 339 },
    { "permil", 8240 },
    { "Yuml", 376 },
    { "scaron", 353 },
    { "Scaron", 352 },
    { "circ", 710 },
    { "tilde", 732 },

    /* Greek letters */
    { "Alpha", 913 },
    { "Beta", 914 },
    { "Gamma", 915 },
    { "Delta", 916 },
    { "Epsilon", 917 },
    { "Zeta", 918 },
    { "Eta", 919 },
    { "Theta", 920 },
    { "Iota", 921 },
    { "Kappa", 922 },
    { "Lambda", 923 },
    { "Mu", 924 },
    { "Nu", 925 },
    { "Xi", 926 },
    { "Omicron", 927 },
    { "Pi", 928 },
    { "Rho", 929 },
    { "Sigma", 931 },
    { "Tau", 932 },
    { "Upsilon", 933 },
    { "Phi", 934 },
    { "Chi", 935 },
    { "Psi", 936 },
    { "Omega", 937 },
    { "alpha", 945 },
    { "beta", 946 },
    { "gamma", 947 },
    { "delta", 948 },
    { "epsilon", 949 },
    { "zeta", 950 },
    { "eta", 951 },
    { "theta", 952 },
    { "iota", 953 },
    { "kappa", 954 },
    { "lambda", 955 },
    { "mu", 956 },
    { "nu", 957 },
    { "xi", 958 },
    { "omicron", 959 },
    { "pi", 960 },
    { "rho", 961 },
    { "sigmaf", 962 },
    { "sigma", 963 },
    { "tau", 964 },
    { "upsilon", 965 },
    { "phi", 966 },
    { "chi", 967 },
    { "psi", 968 },
    { "omega", 969 },
    { "thetasym", 977 },
    { "upsih", 978 },
    { "piv", 982 },

    /* general punctuation marks */
    { "bull", 8226 },
    { "hellip", 8230 },
    { "prime", 8242 },
    { "Prime", 8243 },
    { "oline", 8254 },
    { "frasl", 8260 },

    /* letter-like symbols */
    { "weierp", 8472 },
    { "image", 8465 },
    { "real", 8476 },
    { "alefsym", 8501 },

    /* arrows */
    { "larr", 8592 },
    { "uarr", 8593 },
    { "rarr", 8594 },
    { "darr", 8595 },
    { "harr", 8596 },
    { "crarr", 8629 },
    { "lArr", 8656 },
    { "uArr", 8657 },
    { "rArr", 8658 },
    { "dArr", 8659 },
    { "hArr", 8660 },

    /* mathematical operators */
    { "forall", 8704 },
    { "part", 8706 },
    { "exist", 8707 },
    { "empty", 8709 },
    { "nabla", 8711 },
    { "isin", 8712 },
    { "notin", 8713 },
    { "ni", 8715 },
    { "prod", 8719 },
    { "sum", 8721 },
    { "minus", 8722 },
    { "lowast", 8727 },
    { "radic", 8730 },
    { "prop", 8733 },
    { "infin", 8734 },
    { "ang", 8736 },
    { "and", 8743 },
    { "or", 8744 },
    { "cap", 8745 },
    { "cup", 8746 },
    { "int", 8747 },
    { "there4", 8756 },
    { "sim", 8764 },
    { "cong", 8773 },
    { "asymp", 8776 },
    { "ne", 8800 },
    { "equiv", 8801 },
    { "le", 8804 },
    { "ge", 8805 },
    { "sub", 8834 },
    { "sup", 8835 },
    { "nsub", 8836 },
    { "sube", 8838 },
    { "supe", 8839 },
    { "oplus", 8853 },
    { "otimes", 8855 },
    { "perp", 8869 },
    { "sdot", 8901 },
    { "lceil", 8968 },
    { "rceil", 8969 },
    { "lfloor", 8970 },
    { "rfloor", 8971 },
    { "lang", 9001 },
    { "rang", 9002 },

    /* geometric shapes */
    { "loz", 9674 },

    /* miscellaneous symbols */
    { "spades", 9824 },
    { "clubs", 9827 },
    { "hearts", 9829 },
    { "diams", 9830 },

    /* Latin-extended B */
    { "fnof", 402 },

    /* Latin-2 characters */
    { "Aogon", 260 },
    { "breve", 728 },
    { "Lstrok", 321 },
    { "Lcaron", 317 },
    { "Sacute", 346 },
    { "Scedil", 350 },
    { "Tcaron", 356 },
    { "Zacute", 377 },
    { "Zcaron", 381 },
    { "Zdot", 379 },
    { "aogon", 261 },
    { "ogon", 731 },
    { "lstrok", 322 },
    { "lcaron", 318 },
    { "sacute", 347 },
    { "caron", 711 },
    { "scedil", 351 },
    { "tcaron", 357 },
    { "zacute", 378 },
    { "dblac", 733 },
    { "zcaron", 382 },
    { "zdot", 380 },
    { "Racute", 340 },
    { "Abreve", 258 },
    { "Lacute", 313 },
    { "Cacute", 262 },
    { "Ccaron", 268 },
    { "Eogon", 280 },
    { "Ecaron", 282 },
    { "Dcaron", 270 },
    { "Dstrok", 272 },
    { "Nacute", 323 },
    { "Ncaron", 327 },
    { "Odblac", 336 },
    { "Rcaron", 344 },
    { "Uring", 366 },
    { "Udblac", 368 },
    { "Tcedil", 354 },
    { "racute", 341 },
    { "abreve", 259 },
    { "lacute", 314 },
    { "cacute", 263 },
    { "ccaron", 269 },
    { "eogon", 281 },
    { "ecaron", 283 },
    { "dcaron", 271 },
    { "dstrok", 273 },
    { "nacute", 324 },
    { "ncaron", 328 },
    { "odblac", 337 },
    { "rcaron", 345 },
    { "uring", 367 },
    { "udblac", 369 },
    { "tcedil", 355 },
    { "dot", 729 },
    { 0, 0 }
};


/*
 *   Entity mapping list entry.  We store each entity mapping we find in
 *   the file in one of these structures, and link the structures together
 *   into a list. 
 */
struct entity_map_t
{
    /* length of the expansion */
    size_t exp_len;

    /* local character set expansion */
    unsigned char expansion[1];

    /* 
     *   flag: translation is only for display purposes, and cannot be
     *   used for local-to-Unicode mappings 
     */
    int disp_only : 1;
};

/*
 *   Mapping pages.  Each page maps 256 unicode characters; the page is
 *   selected by the high 8 bits of the unicode code point.  We don't
 *   allocate a page until we set a character on the page for the first
 *   time.
 *   
 *   Each page contains 256 slots for entity_map_t pointers, so each page
 *   is an array of (entity_map_t *) elements.  
 */
static entity_map_t **G_mapping[256];

/*
 *   Construct a unicode character given a page number and index on the
 *   page 
 */
static inline make_unicode(int pagenum, int idx)
{
    return ((unsigned int)pagenum * 256) + (unsigned int)idx;
}

/*
 *   Get the mapping for a particular unicode character value 
 */
static entity_map_t *get_mapping(unsigned int unicode_char)
{
    unsigned int pagenum;

    /* get the page */
    pagenum = ((unicode_char >> 8) & 0xff);

    /* if there's no page allocated here, there's no mapping */
    if (G_mapping[pagenum] == 0)
        return 0;

    /* return the entry at the correct location in this page */
    return G_mapping[pagenum][unicode_char & 0xff];
}

/*
 *   Set the mapping at a given unicode code point.  If there is no page
 *   allocated at the appropriate slot yet, we'll allocate a page here.
 */
static void set_mapping(unsigned int unicode_char, entity_map_t *mapping)
{
    unsigned int pagenum;

    /* get the page number - it's the high 8 bits of the unicode value */
    pagenum = ((unicode_char >> 8) & 0xff);

    /* if the page is not allocated, allocate it now */
    if (G_mapping[pagenum] == 0)
    {
        /* allocate the page and clear it out */
        G_mapping[pagenum] =
            (entity_map_t **)t3malloc(256 * sizeof(entity_map_t *));
        memset(G_mapping[pagenum], 0, 256 * sizeof(entity_map_t *));
    }

    /* set the mapping */
    G_mapping[pagenum][unicode_char & 0xff] = mapping;
}


/*
 *   Read a translation string 
 */
static entity_map_t *read_translation(char *p, char *infile, int linenum)
{
    unsigned char buf[256];
    unsigned char *dstp;
    entity_map_t *mapp;

    /* skip any intervening whitespace */
    for ( ; isspace(*p) ; ++p) ;

    /* read the series of numbers or characters */
    for (dstp = buf ; *p != '\0' ; )
    {
        unsigned int intval;

        /* skip any leading whitespace */
        while (isspace(*p))
            ++p;

        /* if it's the end of the line or a comment, stop */
        if (*p == '\0' || *p == '#')
            break;

        /* if it's a string, read it */
        if (*p == '\'')
        {
            /* scan until the closing quote */
            for (++p ; *p != '\'' && *p != '\0' ; ++p)
            {
                /* if it's an escape, skip it and read the next char */
                if (*p == '\\')
                    ++p;

                /* if we don't have room, abort */
                if (dstp >= buf + sizeof(buf))
                {
                    printf("%s: line %d: entity mapping is too long "
                           "(maximum of %d characters are allowed\n",
                           infile, linenum, sizeof(buf));
                    return 0;
                }

                /* store this character */
                *dstp++ = *p;
            }

            /* if we didn't find a closing quote, it's an error */
            if (*p != '\'')
            {
                printf("%s: line %d: no closing quote found in string\n",
                       infile, linenum);
                return 0;
            }

            /* skip the closing quote */
            ++p;
        }
        else
        {
            size_t clen;
            
            /* scan the character code */
            if (read_number(&intval, &p, infile, linenum, FALSE))
                return 0;

            /* if the number is over 255, it takes two characters */
            clen = (intval > 255 ? 2 : 1);
            
            /* make sure we haven't overflowed our buffer */
            if (dstp + clen > buf + sizeof(buf))
            {
                printf("%s: line %d: entity mapping is too long "
                       "(maximum of %d characters are allowed\n",
                       infile, linenum, sizeof(buf));
                return 0;
            }

            /* add it to the output list */
            *dstp++ = (intval & 0xff);

            /* add the second byte if it's over 255 */
            if (clen > 1)
                *dstp++ = ((intval >> 8) & 0xff);
        }
    }

    /* if we didn't get any characters, it's an error */
    if (dstp == buf)
    {
        printf("%s: line %d: no local character codes found after '='\n",
               infile, linenum);
        return 0;
    }

    /* create a new mapping structure and set it up */
    mapp = (entity_map_t *)t3malloc(sizeof(entity_map_t) + dstp - buf);
    mapp->exp_len = dstp - buf;
    memcpy(mapp->expansion, buf, dstp - buf);

    /* return the mapping */
    return mapp;
}

/*
 *   Parse an entity name mapping 
 */
static entity_map_t *parse_entity(char *p, char *infile, int linenum,
                                  unsigned int *unicode_char)
{
    const char *start;
    const struct entity_t *entp;
    
    /* find the end of the entity name */
    start = p;
    for ( ; isalpha(*p) || isdigit(*p) ; ++p) ;

    /* scan the list for the entity name */
    for (entp = entities ; entp->ename != 0 ; ++entp)
    {
        /* see if this one is a match - note that case is significant */
        if (strlen(entp->ename) == (size_t)(p - start)
            && memcmp(entp->ename, start, p - start) == 0)
            break;
    }

    /* if we didn't find it, it's an error */
    if (entp->ename == 0)
    {
        printf("%s: line %d: unknown entity name \"%.*s\"\n",
               infile, linenum, p - start, start);
        return 0;
    }

    /* tell the caller what the unicode character is */
    *unicode_char = entp->charval;

    /* read and return the translation */
    return read_translation(p, infile, linenum);
}


/*
 *   Main entrypoint 
 */
void main(int argc, char **argv)
{
    int curarg;
    osfildef *fp;
    char *infile;
    char *outfile;
    int linenum;
    int from_unicode_to_local;
    entity_map_t *entp;
    int bidi_count;
    int disp_count;
    unsigned long ofs;
    uchar valbuf[50];
    ulong bidi_bytes;
    ulong disp_bytes;
    int pagenum;
    int i;
    size_t len;
    entity_map_t *default_mapping;

    /* no mappings yet */
    bidi_count = 0;
    disp_count = 0;
    bidi_bytes = 0;
    disp_bytes = 0;
    default_mapping = 0;

    /* presume our mappings are in "local unicode" format */
    from_unicode_to_local = FALSE;

    /* scan options */
    for (curarg = 1 ; curarg < argc && argv[curarg][0] == '-' ; ++curarg)
    {
        /* else */
        {
            /* consume all remaining options so we get a usage message */
            curarg = argc;
            break;
        }
    }

    /* check for required arguments */
    if (curarg + 1 >= argc)
    {
        printf("usage: mkchrtab [options] <source> <dest>\n"
               "  <source> is the input file\n"
               "  <dest> is the output file\n"
              );
        exit(1);
    }

    /* get the input and output filenames */
    infile = argv[curarg];
    outfile = argv[curarg + 1];

    /* open the input file */
    fp = osfoprs(infile, OSFTTEXT);
    if (fp == 0)
    {
        printf("error: unable to open input file \"%s\"\n", infile);
        exit(2);
    }

    /* parse the input file */
    for (linenum = 1 ; ; ++linenum)
    {
        char buf[256];
        char *p;
        unsigned int unicode_char, local_char, extra_char;
        entity_map_t *mapp;

        /* read the next line */
        if (osfgets(buf, sizeof(buf), fp) == 0)
            break;

        /* scan off leading spaces */
        for (p = buf ; *p != '\0' && (isspace(*p) || iscntrl(*p)) ; ++p) ;

        /* if this line is blank, or starts with a '#', ignore it */
        if (*p == '\0' || *p == '\n' || *p == '\r' || *p == '#')
            continue;

        /* check for directives */
        len = strlen(p);
        if (len >= 21
            && memicmp(p, "from_unicode_to_local", 21) == 0
            && (!isalpha(p[21]) && !isdigit(p[21])))
        {
            /* note that we're in display mapping mode */
            from_unicode_to_local = TRUE;

            /* no need to look any further at this line */
            continue;
        }
        else if (len > 15
                 && memicmp(p, "default_display", 15) == 0
                 && isspace(p[15]))
        {
            /* read the character code for the default display */
            mapp = read_translation(p + 16, infile, linenum);

            /* 
             *   if we got a mapping, add it as a display-only mapping
             */
            if (mapp != 0)
            {
                /* mark it display-only */
                mapp->disp_only = TRUE;

                /* add it to the count */
                ++disp_count;
                disp_bytes += 1 + mapp->exp_len;

                /* if we already have a default mapping, warn about it */
                if (default_mapping != 0)
                {
                    printf("%s: line %d: duplicate default_mapping\n",
                           infile, linenum);
                    t3free(default_mapping);
                }

                /* note it as the default mapping */
                default_mapping = mapp;
            }

            /* no need to look any further at this line */
            continue;
        }

        /* check our mode */
        if (from_unicode_to_local)
        {
            /* 
             *   we're doing unicode-to-local translations - these
             *   translations can use SGML entity names for the unicode
             *   characters, and can specify multi-character local
             *   translations 
             */

            /* check for an entity name */
            if (*p == '&')
            {
                /* skip the '&' */
                ++p;
                
                /* parse the entity */
                mapp = parse_entity(p, infile, linenum, &unicode_char);
            }
            else
            {
                /* read the Unicode character number */
                if (read_number(&unicode_char, &p, infile, linenum, FALSE))
                    continue;
                
                /* read the translation */
                mapp = read_translation(p, infile, linenum);
            }

            /* if we created a mapping, mark it as display-only */
            if (mapp != 0)
            {
                /* mark it as display-only */
                mapp->disp_only = TRUE;

                /* count it */
                ++disp_count;

                /* 
                 *   add in the display bytes - count the expansion bytes,
                 *   plus one extra byte for the expansion length prefix
                 *   byte 
                 */
                disp_bytes += 1 + mapp->exp_len;
            }

            /*
             *   If the entry is already set, flag an error and delete the
             *   entry - if it's set bidirectionally, we don't want to set
             *   it in the display-only section 
             */
            if (get_mapping(unicode_char) != 0)
            {
                /* flag the error */
                printf("%s: line %d: mapping previously defined - "
                       "new mapping ignored\n", infile, linenum);
                
                /* delete the mapping and forget it */
                t3free(mapp);
                mapp = 0;
            }
        }
        else
        {
            int local_len;
            
            /*
             *   Doing regular translations.  The format must be two
             *   columns of numbers; the first number is the local
             *   character code, and the second is the unicode character
             *   code.
             *   
             *   Some mapping files contain variations on this format:
             *   
             *   - Some files use three columns, where the first column is
             *   another character set translation (for example, JIS0208
             *   uses the first column to give the Shift-JIS code for each
             *   character).  In these cases, we'll skip the first column,
             *   and use only the last two columns.
             *   
             *   - Some files for multi-byte (i.e., mixed single- and
             *   double-byte characters) have entries for DBCS leading
             *   bytes with no Unicode character specified.  In these
             *   cases, only one column appears on some lines, with DBCS
             *   lead byte value in the local character set, and no
             *   Unicode value at all.  We'll simply ignore these lines
             *   entirely, since they contribute no information that we
             *   need.  
             */

            /* 
             *   read two numbers - the second one is optional, since
             *   we'll skip any single-column entries as explained above 
             */
            if (read_number(&local_char, &p, infile, linenum, FALSE)
                || read_number(&unicode_char, &p, infile, linenum, TRUE))
                continue;

            /* 
             *   check for a third column - if it's present, drop the
             *   first column and keep only the second and third columns 
             */
            if (!read_number(&extra_char, &p, infile, linenum, TRUE))
            {
                /* shift everything over one column */
                local_char = unicode_char;
                unicode_char = extra_char;
            }

            /* 
             *   Create a one-character mapping.  If the local character
             *   is over 255, use two bytes for the local mapping,
             *   otherwise use a single byte. 
             */
            local_len = (local_char > 255 ? 2 : 1);
            mapp = (entity_map_t *)
                   t3malloc(sizeof(entity_map_t) + local_len);
            mapp->exp_len = local_len;

            /* set the first byte of the expansion to the low-order byte */
            mapp->expansion[0] = (local_char & 0xff);

            /* 
             *   set the second byte of the expansion to the high-order
             *   byte, if indeed there is a second byte 
             */
            if (local_len > 1)
                mapp->expansion[1] = ((local_char >> 8) & 0xff);

            /* mappings created this way are reversible */
            mapp->disp_only = FALSE;

            /* count the mapping */
            ++bidi_count;
            bidi_bytes += local_len;
        }

        /* add the mapping into our mapping list */
        if (mapp != 0)
            set_mapping(unicode_char, mapp);
    }

    /* we're done with the input file */
    osfcls(fp);

    /* open the output file */
    fp = osfopwb(outfile, OSFTCMAP);
    if (fp == 0)
    {
        printf("error: unable to open output file \"%s\"\n", outfile);
        exit(2);
    }

    /*
     *   Calculate where the unicode-to-local display mapping appears.
     *   This mapping immediately follows the local-to-unicode
     *   bidirectional mapping; each entry in the bidirectional mapping is
     *   of a fixed size, so we can easily figure out how much space it
     *   will take up.
     *   
     *   Each bidirectional record takes up 4 bytes (two UINT2's)
     *.  A UINT4 is needed for this header itself.
     *.  The bidirectional section contains at least one UINT2 (for its
     *   own count header) 
     */
    ofs = (bidi_count * 4) + 4 + 2;
    oswp4(valbuf, ofs);

    /* add in the count of bidirectional records */
    oswp2(valbuf + 4, bidi_count);

    /* write out the header */
    if (osfwb(fp, valbuf, 6))
    {
        printf("error writing header\n");
        goto done;
    }

    /* write the bidirectional records */
    for (pagenum = 0 ; pagenum < 256 ; ++pagenum)
    {
        /* if there's no page here, skip it */
        if (G_mapping[pagenum] == 0)
            continue;

        /* go through the mappings in this page */
        for (i = 0 ; i < 256 ; ++i)
        {
            /* 
             *   if this character is mapped and it's bidirectional, write
             *   it out 
             */
            entp = G_mapping[pagenum][i];
            if (entp != 0 && !entp->disp_only)
            {
                /* 
                 *   the entry has a UINT2 for the unicode value, followed
                 *   by a UINT2 for the local value 
                 */
                oswp2(valbuf, make_unicode(pagenum, i));
                
                /* 
                 *   if the translation is one byte long, fill in zero in
                 *   the high part; otherwise, fill in both bytes 
                 */
                valbuf[2] = entp->expansion[0];
                valbuf[3] = (entp->exp_len > 1 ? entp->expansion[1] : 0);
                
                /* write it out */
                if (osfwb(fp, valbuf, 4))
                {
                    printf("error writing mapping to file\n");
                    goto done;
                }
            }
        }
    }

    /* 
     *   Write out the number of display mappings, and the number of bytes
     *   of display mappings.  Note that we can use both the bidirectional
     *   mappings and the display-only mappings for unicode-to-local
     *   mappings, so write out the number that we have of both types.
     *   The byte size of the mappings must include the length prefixes;
     *   there's one length prefix byte per entry.  
     */
    oswp2(valbuf, disp_count + bidi_count);
    oswp4(valbuf + 2, disp_bytes + bidi_bytes + disp_count + bidi_count);
    if (osfwb(fp, valbuf, 6))
    {
        printf("error writing display translation header\n");
        goto done;
    }

    /* check for a default display mapping */
    if (default_mapping == 0)
    {
        printf("error: no 'default_display' mapping specified\n");
        goto done;
    }
    else
    {
        /* write out the default display entity */
        oswp2(valbuf, 0);
        valbuf[2] = (uchar)default_mapping->exp_len;
        if (osfwb(fp, valbuf, 3)
            || osfwb(fp, default_mapping->expansion,
                     default_mapping->exp_len))
        {
            printf("error writing default_display mapping to output file\n");
            goto done;
        }
    }

    /* 
     *   write all of the mappings - both the display and bidirectional
     *   mappings provide unicode-to-local mappings 
     */
    /* go through all of the pages */
    for (pagenum = 0 ; pagenum < 256 ; ++pagenum)
    {
        /* if this page isn't present, skip it */
        if (G_mapping[pagenum] == 0)
            continue;
        
        /* go through the page */
        for (i = 0 ; i < 256 ; ++i)
        {
            /* if this entry exists, write it out */
            entp = G_mapping[pagenum][i];
            if (entp != 0)
            {
                /* write out this entity */
                oswp2(valbuf, make_unicode(pagenum, i));
                valbuf[2] = (uchar)entp->exp_len;
                if (osfwb(fp, valbuf, 3)
                    || osfwb(fp, entp->expansion, entp->exp_len))
                {
                    printf("error writing display mapping "
                           "to output file\n");
                    break;
                }
            }
        }
    }

    /* write out the end-of-file marker */
    if (osfwb(fp, "$EOF", 4))
        printf("error writing end-of-file marker\n");

done:
    /* delete the entity list */
    for (pagenum = 0 ; pagenum < 256 ; ++pagenum)
    {
        /* if there's no mapping here, ignore it */
        if (G_mapping[pagenum] == 0)
            continue;

        /* delete each entry on the page */
        for (i = 0 ; i < 256 ; ++i)
        {
            if (G_mapping[pagenum][i] != 0)
                t3free(G_mapping[pagenum][i]);
        }

        /* delete the page itself */
        t3free(G_mapping[pagenum]);
    }

    /* delete the default mapping, if any */
    if (default_mapping != 0)
        t3free(default_mapping);
    
    /* done with the output file */
    osfcls(fp);

    /* success */
    exit(0);
}

