#ifdef RCSID
static char RCSid[] =
"$Header$";
#endif

/* Copyright (c) 1999 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  tcmakecl.cpp - TADS Compiler "Make" command line tool
Function
  Command-line interface to "make" program build utility
Notes
  
Modified
  07/11/99 MJRoberts  - Creation
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "os.h"
#include "t3std.h"
#include "tcmake.h"
#include "tchostsi.h"
#include "vmerr.h"
#include "tcvsn.h"
#include "resload.h"
#include "tcmain.h"


/* get an option argument */
static char *get_opt_arg(int argc, char **argv, int *curarg, int optlen)
{
    /* 
     *   if it's jammed up against the option letter, get it from the
     *   current array entry; otherwise, get it from the next array entry 
     */
    if (argv[*curarg][optlen + 1] != '\0')
    {
        /* it's attached to the same option entry */
        return argv[*curarg] + optlen + 1;
    }
    else if (*curarg + 1 < argc)
    {
        /* it's by itself as the next option entry */
        ++(*curarg);
        return argv[*curarg];
    }
    else
    {
        /* it's not present at all */
        return 0;
    }
}

/*
 *   Parse an options file.  If argv is null, we'll simply count the
 *   arguments and return the count.  If argv is non-null, it must be
 *   allocated with the proper number of slots for the number of arguments
 *   in the file (as obtained from a call to this function with argv =
 *   null); we'll allocate argv[i] for each i>0 using lib_copy_str().  
 */
static int parse_opt_file(osfildef *fp, char **argv)
{
    char *buf;
    size_t buflen;
    int argc;

    /* allocate our initial buffer */
    buflen = 512;
    buf = (char *)t3malloc(buflen);
    
    /* keep going until we run out of text in the file */
    for (argc = 1 ; ; )
    {
        size_t len;
        char *p;
        
        /* read the next line - stop if we're done */
        if (osfgets(buf, buflen, fp) == 0)
            break;

        /* check for proper termination */
        for (;;)
        {
            /* get the buffer length */
            len = strlen(buf);

            /* check for a newline of some kind at the end */
            if (len != 0 && (buf[len - 1] == '\n' || buf[len - 1] == '\r'))
            {
                /* 
                 *   remove consecutive newline/return characters - in
                 *   case we are using a file moved from another system
                 *   with incompatible newline conventions, simply remove
                 *   all of the newlines we find 
                 */
                while (len != 0
                       && (buf[len-1] == '\n' || buf[len - 1] == '\r'))
                    --len;

                /* null-terminate the line */
                buf[len] = '\0';

                /* we now have our line */
                break;
            }
            else if (len == buflen - 1)
            {
                /* 
                 *   The buffer was completely filled, and wasn't
                 *   null-terminated - the line must be too long for the
                 *   buffer.  Expand the buffer and go back for more.  
                 */
                buflen += 512;
                buf = (char *)t3realloc(buf, buflen);

                /* append more data into the line buffer and try again */
                if (osfgets(buf + len, buflen - len, fp) == 0)
                    break;
            }
        }

        /* parse the options on this line */
        for (p = buf ; *p != '\0' ; )
        {
            char *start;
            size_t arg_len;

            /* skip leading spaces */
            for ( ; is_space(*p) ; ++p) ;

            /* 
             *   if we've reached the end of the line or a command
             *   character ('#'), we're done with this line 
             */
            if (*p == '\0' || *p == '#')
                break;

            /* 
             *   if the argument is quoted, find the matching quote;
             *   otherwise, look for the next space 
             */
            if (*p == '"')
            {
                char *dst;
                
                /* find the matching quote */
                for (++p, start = dst = p ; *p != '\0' ; ++p)
                {
                    /* 
                     *   if this is a quote, check for stuttering - if
                     *   it's stuttered, treat it as a single quote (and
                     *   convert it to same), otherwise we've reached the
                     *   end of the argument 
                     */
                    if (*p == '"')
                    {
                        /* check for stuttering */
                        if (*(p+1) == '"')
                        {
                            /* 
                             *   it's stuttered - skip the extra quote, so
                             *   that we copy only a single quote 
                             */
                            ++p;
                        }
                        else
                        {
                            /* this is our closing quote - skip it */
                            ++p;

                            /* we're done with the argument */
                            break;
                        }
                    }

                    /* copy this character to the output */
                    *dst++ = *p;
                }

                /* note the length of the argument */
                arg_len = dst - start;
            }
            else
            {
                /* find the next space */
                for (start = p ; *p != '\0' && !is_space(*p) ; ++p) ;

                /* note the length of the argument */
                arg_len = p - start;
            }

            /* store the argument if we have an output vector */
            if (argv != 0)
                argv[argc] = lib_copy_str(start, arg_len);

            /* count the argument */
            ++argc;
        }
    }

    /* done with our line buffer - delete it */
    t3free(buf);

    /* return the argument count */
    return argc;
}

/*
 *   Program main entrypoint 
 */
void main(int argc, char **argv)
{
    CTcHostIfcStdio hostifc;
    int curarg;
    char *p;
    int force_build = FALSE;
    int force_link = FALSE;
    CTcMake *mk;
    int err_cnt, warn_cnt;
    int image_specified = FALSE;
    char dirbuf[OSFNMAX];
    int show_banner;
    char *opt_file = 0;
    int read_opt_file = FALSE;
    int usage_err = FALSE;
    int add_def_mod = TRUE;
    int compile_only = FALSE;
    osfildef *string_fp = 0;

    /* initialize the error subsystem */
    {
        CResLoader *res_loader;
        char exe_path[OSFNMAX];

        /* create a resource loader */
        os_get_path_name(exe_path, sizeof(exe_path), argv[0]);
        res_loader = new CResLoader(exe_path);
        res_loader->set_exe_filename(argv[0]);

        /* initialize the error subsystem */
        CTcMain::tc_err_init(1024, res_loader);

        /* done with the resource loader */
        delete res_loader;
    }

    /* create the 'make' object */
    mk = new CTcMake();

    /* presume we'll show the banner */
    show_banner = TRUE;

parse_options:
    /* read the options */
    for (curarg = 1 ; curarg < argc && argv[curarg][0] == '-' ; ++curarg)
    {
        switch(argv[curarg][1])
        {
        case 'f':
            /* read an options file - make sure we don't have one already */
            if (opt_file != 0)
            {
                /* can't read a file from a file */
                printf("error: only one option file (-f) is allowed\n");
                err_cnt = 1;
                goto done;
            }

            /* 
             *   remember the options file name - we'll get around to the
             *   actual reading of the file after we finish with the
             *   command line options 
             */
            opt_file = get_opt_arg(argc, argv, &curarg, 1);
            if (opt_file == 0)
                goto bad_option;
            break;
            
        case 'd':
            /* set debug mode */
            mk->set_debug(TRUE);
            break;

        case 'D':
            /* add preprocessor symbol definition */
            p = get_opt_arg(argc, argv, &curarg, 1);
            if (p != 0)
            {
                char *eq;

                /* see if there's an '=' in the string */
                for (eq = p ; *eq != '\0' && *eq != '=' ; ++eq) ;

                /* if the '=' is present, replace it with a null byte */
                if (*eq == '=')
                {
                    /* replace it */
                    *eq = '\0';

                    /* skip to the start of the replacement text */
                    ++eq;
                }

                /* add the symbol definition */
                mk->def_pp_sym(p, eq);
            }
            else
                goto bad_option;
            break;

        case 'U':
            /* add preprocess symbol un-definition */
            p = get_opt_arg(argc, argv, &curarg, 1);
            if (p != 0)
                mk->undef_pp_sym(p);
            else
                goto bad_option;
            break;

        case 'c':
            /* see what follows */
            switch (argv[curarg][2])
            {
            case 's':
                /* it's a character set specification */
                p = get_opt_arg(argc, argv, &curarg, 2);
                if (p != 0)
                    mk->set_source_charset(p);
                else
                    goto bad_option;
                break;

            case '\0':
                /* set compile-only (no link) mode */
                mk->set_do_link(FALSE);
                compile_only = TRUE;
                break;

            default:
                /* invalid option */
                goto bad_option;
            }
            break;

        case 'r':
            /* see what follows */
            switch(argv[curarg][2])
            {
            case 'l':
                /* force only the link phase */
                force_link = TRUE;
                break;

            case '\0':
                force_build = TRUE;
                break;

            default:
                /* invalid option */
                goto bad_option;
            }
            break;
            
        case 'v':
            /* set verbose mode */
            mk->set_verbose(TRUE);
            break;
            
        case 'I':
            /* add a #include path */
            p = get_opt_arg(argc, argv, &curarg, 1);
            if (p != 0)
                mk->add_include_path(p);
            else
                goto bad_option;
            break;
            
        case 'o':
            /* set the image file name */
            p = get_opt_arg(argc, argv, &curarg, 1);
            if (p != 0)
            {
                /* set the image file name */
                mk->set_image_file(p);

                /* note that we have an explicit image file name */
                image_specified = TRUE;
            }
            else
                goto bad_option;
            break;

        case 'O':
            switch(argv[curarg][2])
            {
            case 's':
                /* 
                 *   if we already have a string capture file, close the
                 *   old one 
                 */
                if (string_fp != 0)
                {
                    osfcls(string_fp);
                    string_fp = 0;
                }
                
                /* set the string file */
                p = get_opt_arg(argc, argv, &curarg, 2);
                if (p == 0)
                    goto bad_option;

                /* open the string file */
                string_fp = osfopwt(p, OSFTTEXT);
                if (string_fp == 0)
                {
                    printf("error: unable to create string capture file "
                           "\"%s\"\n", p);
                    goto done;
                }

                /* set the string capture file in the build object */
                mk->set_string_capture(string_fp);

                /* done */
                break;

            default:
                /* invalid suboption */
                goto bad_option;
            }
            break;

        case 'F':
            /* output path settings - see which one we're setting */
            switch(argv[curarg][2])
            {
            case 's':
                /* set the source path */
                p = get_opt_arg(argc, argv, &curarg, 2);
                if (p != 0)
                    mk->set_source_dir(p);
                else
                    goto bad_option;
                break;
                
            case 'y':
                /* set the symbol path */
                p = get_opt_arg(argc, argv, &curarg, 2);
                if (p != 0)
                    mk->set_symbol_dir(p);
                else
                    goto bad_option;
                break;
            
            case 'o':
                /* set the object path */
                p = get_opt_arg(argc, argv, &curarg, 2);
                if (p != 0)
                    mk->set_object_dir(p);
                else
                    goto bad_option;
                break;

            default:
                /* invalid option */
                goto bad_option;
            }
            break;

        case 'n':
            /* check what we have */
            if (strcmp(argv[curarg], "-nopre") == 0)
            {
                /* explicitly turn off preinit mode */
                mk->set_preinit(FALSE);
            }
            else if (strcmp(argv[curarg], "-nobanner") == 0)
            {
                /* turn off the banner */
                show_banner = FALSE;
            }
            else if (strcmp(argv[curarg], "-nodef") == 0)
            {
                /* don't add the default modules */
                add_def_mod = FALSE;
            }
            else
                goto bad_option;
            break;

        case 'p':
            /* check what we have */
            if (strcmp(argv[curarg], "-pre") == 0)
            {
                /* explicitly turn on preinit mode */
                mk->set_preinit(TRUE);
            }
            else
                goto bad_option;
            break;

        case 'w':
            /* warning level - see what level they're setting */
            switch(argv[curarg][2])
            {
            case '0':
                /* no warnings */
                mk->set_warnings(FALSE);
                mk->set_pedantic(FALSE);
                break;

            case '1':
                /* normal warnings, but not pedantic warnings */
                mk->set_warnings(TRUE);
                mk->set_pedantic(FALSE);
                break;

            case '2':
                /* show all warnings, even pedantic ones */
                mk->set_warnings(TRUE);
                mk->set_pedantic(TRUE);
                break;

            default:
                /* invalid option */
                goto bad_option;
            }
            break;

        case '?':
            /* take '?' as an explicit request to show the usage message */
            goto bad_option;
            
        default:
        bad_option:
            /* invalid - skip all remaining arguments so we show usage */
            curarg = argc;
            usage_err = TRUE;
            break;
        }
    }

    /* 
     *   if there are no module arguments, and no option file was
     *   specified, try reading "makefile.t3m" by default 
     */
    if (curarg >= argc && !usage_err && opt_file == 0)
    {
        /* 
         *   if "makefile.t3m" exists, try reading it; otherwise, show the
         *   usage error 
         */
        if (!osfacc("makefile.t3m"))
            opt_file = "makefile.t3m";
        else
            usage_err = TRUE;
    }

    /* 
     *   Show the banner, unless they explicitly asked us not to (and even
     *   then, show it anyway if we're going to show the "usage" message).
     */
    if ((show_banner || usage_err) && (opt_file == 0 || read_opt_file))
        printf("TADS Compiler %d.%d.%d  "
               "Copyright 1999, 2000 Michael J. Roberts\n",
               TC_VSN_MAJOR, TC_VSN_MINOR, TC_VSN_PATCH);

    /* if a usage error occurred, display the usage message */
    if (usage_err)
    {
        printf("usage: t3make [options] module ...\n"
               "options:\n"
               "  -c      - compile only (do not link image file)\n"
               "  -cs xxx - source file character set is 'xxx'\n"
               "  -d      - compile for debugging (include symbols)\n"
               "  -D x=y  - define preprocessor symbol 'x' with value 'y'\n"
               "  -f file - read command line options from 'file'\n"
               "  -I dir  - add 'dir' to #include search path\n"
               "  -Fs dir - look for source files in directory 'dir'\n"
               "  -Fy dir - put symbol files in directory 'dir'\n"
               "  -Fo dir - put object files in directory 'dir'\n"
               "  -nobanner - suppress version/copyright banner\n"
               "  -nopre  - do not run preinit regardless of debug mode\n"
               "  -o img  - set image filename to 'img'\n"
               "  -Os str - write all strings found in compiled files to "
               "text file 'str'\n"
               "  -pre    - run preinit regardless of debug mode\n"
               "  -r      - recompile all object files, "
               "even if not out of date\n"
               "  -rl     - relink, even if image is not out of date\n"
               "  -U x    - un-define preprocessor symbol 'x'\n"
               "  -v      - display verbose error messages\n"
               "  -w#     - warning level: 0=none, 1=standard "
               "(default), 2=pedantic\n"
               "\n"
               "Each 'module' is the name of a source file to compile, "
               "with the default suffix\n"
               "'.t'.  For each source file, the object file has the same "
               "name with the suffix\n"
               "'.t3o', and the symbol file has suffix '.t3s'.  The source, "
               "object, and symbol\n"
               "files will be in the same directory unless other directories "
               "are specified\n"
               "with -Fs, -Fo, and/or -Fy options.\n"
               "\n"
               "If no modules are specified, we will attempt to find "
               "'makefile.t3m' in the\ncurrent directory, and read it to "
               "obtain build instructions.\n");

        /* terminate with error */
        err_cnt = 1;
        goto done;
    }

    /* add the modules */
    for ( ; curarg < argc ; ++curarg)
    {
        /* add this module */
        mk->add_module(argv[curarg], 0, 0);

        /* 
         *   if no image was specified, use the first module's name as the
         *   image name, with the suffix ".t3x" 
         */
        if (!image_specified)
        {
            char buf[OSFNMAX];

            /* build the default image filename */
            strcpy(buf, argv[curarg]);
            os_remext(buf);
            os_addext(buf, "t3x");

            /* set the filename */
            mk->set_image_file(buf);

            /* 
             *   note that we've set the name, so we don't try to set it
             *   again 
             */
            image_specified = TRUE;
        }
    }

    /*
     *   If an option file was specified or implied, and we haven't read
     *   it yet, read it now.  
     */
    if (opt_file != 0 && !read_opt_file)
    {
        osfildef *fp;
        int new_argc;
        char **new_argv;
        char new_opt_file[OSFNMAX];

        /* if the options file doesn't exist, try adding the suffix 't3m' */
        if (osfacc(opt_file))
        {
            strcpy(new_opt_file, opt_file);
            os_defext(new_opt_file, "t3m");
            opt_file = new_opt_file;
        }

        /* open the options file */
        fp = osfoprt(opt_file, OSFTTEXT);
        if (fp == 0)
        {
            printf("error: unable to read option file \"%s\"\n", opt_file);
            err_cnt = 1;
            goto done;
        }

        /* first, parse the file simply to count the arguments */
        new_argc = parse_opt_file(fp, 0);

        /* rewind the file to parse it again */
        osfseek(fp, 0, OSFSK_SET);

        /* allocate a new argument list */
        new_argv = (char **)t3malloc(new_argc * sizeof(new_argv[0]));

        /* copy the program name from argv[0] */
        new_argv[0] = argv[0];

        /* re-read the file, saving the arguments */
        parse_opt_file(fp, new_argv);

        /* done with the file - close it */
        osfcls(fp);

        /* start over with the new argument vector */
        argv = new_argv;
        argc = new_argc;
        
        /* 
         *   note that we've read the options file - we can only do this
         *   once per run 
         */
        read_opt_file = TRUE;

        /* go add the new command options */
        goto parse_options;
    }

    /* 
     *   Add the default object modules, if appropriate.  Do not add the
     *   default modules unless we're linking. 
     */
    if (add_def_mod && !compile_only)
    {
        char exepath[OSFNMAX];
        char outpath[OSFNMAX];
        char srcname[OSFNMAX];
        char symname[OSFNMAX];
        char objname[OSFNMAX];

        /* 
         *   get the compiler's path - the default modules are in the
         *   compiler's directory 
         */
        os_get_path_name(exepath, sizeof(exepath), argv[0]);

        /* 
         *   get the image file directory - we'll put the derived files
         *   for the default modules into this directory by default 
         */
        os_get_path_name(outpath, sizeof(outpath), mk->get_image_file());

        /* add the "_main" module */
        os_build_full_path(srcname, sizeof(srcname), exepath, "_main.t");
        os_build_full_path(symname, sizeof(symname), outpath, "_main.t3s");
        os_build_full_path(objname, sizeof(objname), outpath, "_main.t3o");
        mk->add_module_first(srcname, symname, objname);
    }

    /* 
     *   add the compiler's own directory to the include path - add this
     *   now so that it's the last directory in the path, so that we
     *   always look in any user-specified directories before we look in
     *   the compile's directory 
     */
    os_get_path_name(dirbuf, sizeof(dirbuf), argv[0]);
    mk->add_include_path(dirbuf);

    /* build the program */
    err_cnt = warn_cnt = 0;
    mk->build(&hostifc, &err_cnt, &warn_cnt,
              force_build, force_link, argv[0]);

done:
    /* terminate the error subsystem */
    CTcMain::tc_err_term();

    /* delete our 'make' object */
    delete mk;

    /* if we read an options file, delete the memory used for the options */
    if (read_opt_file)
    {
        int i;
        
        /* 
         *   delete the argv strings (except for argv[0], which came from
         *   the caller)
         */
        for (i = 1 ; i < argc ; ++i)
            lib_free_str(argv[i]);

        /* delete the argv vector itself */
        t3free(argv);
    }

    /* if we have a string capture file, close it */
    if (string_fp != 0)
        osfcls(string_fp);

    /* show any unfreed memory (if we're in a debug build) */
    t3_list_memory_blocks();

    /* 
     *   exit with an appropriate exit code, depending on whether we had
     *   any errors or not 
     */
    exit(err_cnt != 0 ? OSEXFAIL : OSEXSUCC);
}

