#ifdef RCSID
static char RCSid[] =
"$Header$";
#endif

/* Copyright (c) 1987, 1999 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmconsol.cpp - TADS 3 console input reader and output formatter
Function
  Provides console input and output for the TADS 3 built-in function set
  for the T3 VM, including the output formatter.

  T3 uses the UTF-8 character set to represent character strings.  The OS
  functions use the local character set.  We perform the mapping between
  UTF-8 and the local character set within this module, so that OS routines
  see local characters only, not UTF-8.

  This code is based on the TADS 2 output formatter, but has been
  substantially reworked for C++, Unicode, and the slightly different
  TADS 3 formatting model.
Notes

Returns
  None
Modified
  08/25/99 MJRoberts  - created from TADS 2 output formatter
*/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <wchar.h>

#include "os.h"
#include "t3std.h"
#include "utf8.h"
#include "charmap.h"
#include "vmuni.h"
#include "vmconsol.h"
#include "vmglob.h"


/* ------------------------------------------------------------------------ */
/*
 *   Case-insensitive wide-string comparison routine.  strcmp() semantics.
 */
static int w_stricmp(const wchar_t *a, const wchar_t *b)
{
    /* loop over characters until we find a mismatch */
    for ( ; ; ++a, ++b)
    {
        wchar_t ua, ub;

        /* get both characters converted to upper-case */
        ua = t3_to_upper(*a);
        ub = t3_to_upper(*b);

        /* if they're different, return the appropriate sense indicator */
        if (ua > ub)
            return 1;
        else if (ua < ub)
            return -1;

        /* 
         *   if they both ended here, the strings are equal (if one ended,
         *   the other must have as well, since we only made it this far
         *   if the two characters were equal) 
         */
        if (*a == L'\0')
            return 0;
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Log-file formatter subclass implementation 
 */

/*
 *   Open a new log file 
 */
int CVmFormatterLog::open_log_file(const char *fname, int initial_html_mode)
{
    /* close any existing log file */
    if (close_log_file())
        return 1;
    
    /* reinitialize */
    init();

    /* save the filename for later (we'll need it when we close the file) */
    strcpy(logfname_, fname);

    /* open the new file */
    logfp_ = osfopwt(fname, OSFTLOG);

    /* set our initial HTML mode */
    html_mode_ = initial_html_mode;

    /* return success if we successfully opened the file, failure otherwise */
    return (logfp_ == 0);
}

/*
 *   Close the log file 
 */
int CVmFormatterLog::close_log_file()
{
    /* if we have a file, close it */
    if (logfp_ != 0)
    {
        /* close the handle */
        osfcls(logfp_);

        /* set the system file type to "log file" */
        os_settype(logfname_, OSFTLOG);

        /* forget about our log file handle */
        logfp_ = 0;
    }

    /* success */
    return 0;
}


/* ------------------------------------------------------------------------ */
/*
 *   HTML entity mapping table.  When we're in non-HTML mode, we keep our
 *   own expansion table so that we can map HTML entity names into the
 *   local character set.
 *   
 *   The entries in this table must be in sorted order (by HTML entity
 *   name), because we use a binary search to find an entity name in the
 *   table.  
 */

/*
 *   '&' character name definition structure 
 */
struct amp_tbl_t
{
    /* entity name */
    const char *cname;

    /* HTML Unicode character value */
    uint        html_cval;
};

/* 
 *   table of '&' character name sequences 
 */
static const struct amp_tbl_t amp_tbl[] =
{
    { "AElig", 198 },
    { "Aacute", 193 },
    { "Abreve", 258 },
    { "Acirc", 194 },
    { "Agrave", 192 },
    { "Alpha", 913 },
    { "Aogon", 260 },
    { "Aring", 197 },
    { "Atilde", 195 },
    { "Auml", 196 },
    { "Beta", 914 },
    { "Cacute", 262 },
    { "Ccaron", 268 },
    { "Ccedil", 199 },
    { "Chi", 935 },
    { "Dagger", 8225 },
    { "Dcaron", 270 },
    { "Delta", 916 },
    { "Dstrok", 272 },
    { "ETH", 208 },
    { "Eacute", 201 },
    { "Ecaron", 282 },
    { "Ecirc", 202 },
    { "Egrave", 200 },
    { "Eogon", 280 },
    { "Epsilon", 917 },
    { "Eta", 919 },
    { "Euml", 203 },
    { "Gamma", 915 },
    { "Iacute", 205 },
    { "Icirc", 206 },
    { "Igrave", 204 },
    { "Iota", 921 },
    { "Iuml", 207 },
    { "Kappa", 922 },
    { "Lacute", 313 },
    { "Lambda", 923 },
    { "Lcaron", 317 },
    { "Lstrok", 321 },
    { "Mu", 924 },
    { "Nacute", 323 },
    { "Ncaron", 327 },
    { "Ntilde", 209 },
    { "Nu", 925 },
    { "OElig", 338 },
    { "Oacute", 211 },
    { "Ocirc", 212 },
    { "Odblac", 336 },
    { "Ograve", 210 },
    { "Omega", 937 },
    { "Omicron", 927 },
    { "Oslash", 216 },
    { "Otilde", 213 },
    { "Ouml", 214 },
    { "Phi", 934 },
    { "Pi", 928 },
    { "Prime", 8243 },
    { "Psi", 936 },
    { "Racute", 340 },
    { "Rcaron", 344 },
    { "Rho", 929 },
    { "Sacute", 346 },
    { "Scaron", 352 },
    { "Scedil", 350 },
    { "Sigma", 931 },
    { "THORN", 222 },
    { "Tau", 932 },
    { "Tcaron", 356 },
    { "Tcedil", 354 },
    { "Theta", 920 },
    { "Uacute", 218 },
    { "Ucirc", 219 },
    { "Udblac", 368 },
    { "Ugrave", 217 },
    { "Upsilon", 933 },
    { "Uring", 366 },
    { "Uuml", 220 },
    { "Xi", 926 },
    { "Yacute", 221 },
    { "Yuml", 376 },
    { "Zacute", 377 },
    { "Zcaron", 381 },
    { "Zdot", 379 },
    { "Zeta", 918 },
    { "aacute", 225 },
    { "abreve", 259 },
    { "acirc", 226 },
    { "acute", 180 },
    { "aelig", 230 },
    { "agrave", 224 },
    { "alefsym", 8501 },
    { "alpha", 945 },
    { "amp", '&' },
    { "and", 8743 },
    { "ang", 8736 },
    { "aogon", 261 },
    { "aring", 229 },
    { "asymp", 8776 },
    { "atilde", 227 },
    { "auml", 228 },
    { "bdquo", 8222 },
    { "beta", 946 },
    { "breve", 728 },
    { "brvbar", 166 },
    { "bull", 8226 },
    { "cacute", 263 },
    { "cap", 8745 },
    { "caron", 711 },
    { "ccaron", 269 },
    { "ccedil", 231 },
    { "cedil", 184 },
    { "cent", 162 },
    { "chi", 967 },
    { "circ", 710 },
    { "clubs", 9827 },
    { "cong", 8773 },
    { "copy", 169 },
    { "crarr", 8629 },
    { "cup", 8746 },
    { "curren", 164 },
    { "dArr", 8659 },
    { "dagger", 8224 },
    { "darr", 8595 },
    { "dblac", 733 },
    { "dcaron", 271 },
    { "deg", 176 },
    { "delta", 948 },
    { "diams", 9830 },
    { "divide", 247 },
    { "dot", 729 },
    { "dstrok", 273 },
    { "eacute", 233 },
    { "ecaron", 283 },
    { "ecirc", 234 },
    { "egrave", 232 },
    { "emdash", 8212 },
    { "empty", 8709 },
    { "endash", 8211 },
    { "eogon", 281 },
    { "epsilon", 949 },
    { "equiv", 8801 },
    { "eta", 951 },
    { "eth", 240 },
    { "euml", 235 },
    { "exist", 8707 },
    { "fnof", 402 },
    { "forall", 8704 },
    { "frac12", 189 },
    { "frac14", 188 },
    { "frac34", 190 },
    { "frasl", 8260 },
    { "gamma", 947 },
    { "ge", 8805 },
    { "gt", '>' },
    { "hArr", 8660 },
    { "harr", 8596 },
    { "hearts", 9829 },
    { "hellip", 8230 },
    { "iacute", 237 },
    { "icirc", 238 },
    { "iexcl", 161 },
    { "igrave", 236 },
    { "image", 8465 },
    { "infin", 8734 },
    { "int", 8747 },
    { "iota", 953 },
    { "iquest", 191 },
    { "isin", 8712 },
    { "iuml", 239 },
    { "kappa", 954 },
    { "lArr", 8656 },
    { "lacute", 314 },
    { "lambda", 955 },
    { "lang", 9001 },
    { "laquo", 171 },
    { "larr", 8592 },
    { "lcaron", 318 },
    { "lceil", 8968 },
    { "ldq", 8220 },
    { "ldquo", 8220 },
    { "le", 8804 },
    { "lfloor", 8970 },
    { "lowast", 8727 },
    { "loz", 9674 },
    { "lsaquo", 8249 },
    { "lsq", 8216 },
    { "lsquo", 8216 },
    { "lstrok", 322 },
    { "lt", '<' },
    { "macr", 175 },
    { "mdash", 8212 },
    { "micro", 181 },
    { "middot", 183 },
    { "minus", 8722 },
    { "mu", 956 },
    { "nabla", 8711 },
    { "nacute", 324 },
    { "nbsp", QSPACE },
    { "ncaron", 328 },
    { "ndash", 8211 },
    { "ne", 8800 },
    { "ni", 8715 },
    { "not", 172 },
    { "notin", 8713 },
    { "nsub", 8836 },
    { "ntilde", 241 },
    { "nu", 957 },
    { "oacute", 243 },
    { "ocirc", 244 },
    { "odblac", 337 },
    { "oelig", 339 },
    { "ogon", 731 },
    { "ograve", 242 },
    { "oline", 8254 },
    { "omega", 969 },
    { "omicron", 959 },
    { "oplus", 8853 },
    { "or", 8744 },
    { "ordf", 170 },
    { "ordm", 186 },
    { "oslash", 248 },
    { "otilde", 245 },
    { "otimes", 8855 },
    { "ouml", 246 },
    { "para", 182 },
    { "part", 8706 },
    { "permil", 8240 },
    { "perp", 8869 },
    { "phi", 966 },
    { "pi", 960 },
    { "piv", 982 },
    { "plusmn", 177 },
    { "pound", 163 },
    { "prime", 8242 },
    { "prod", 8719 },
    { "prop", 8733 },
    { "psi", 968 },
    { "quot", '"' },
    { "rArr", 8658 },
    { "racute", 341 },
    { "radic", 8730 },
    { "rang", 9002 },
    { "raquo", 187 },
    { "rarr", 8594 },
    { "rcaron", 345 },
    { "rceil", 8969 },
    { "rdq", 8221 },
    { "rdquo", 8221 },
    { "real", 8476 },
    { "reg", 174 },
    { "rfloor", 8971 },
    { "rho", 961 },
    { "rsaquo", 8250 },
    { "rsq", 8217 },
    { "rsquo", 8217 },
    { "sacute", 347 },
    { "sbquo", 8218 },
    { "scaron", 353 },
    { "scedil", 351 },
    { "sdot", 8901 },
    { "sect", 167 },
    { "shy", 173 },
    { "sigma", 963 },
    { "sigmaf", 962 },
    { "sim", 8764 },
    { "spades", 9824 },
    { "sub", 8834 },
    { "sube", 8838 },
    { "sum", 8721 },
    { "sup", 8835 },
    { "sup1", 185 },
    { "sup2", 178 },
    { "sup3", 179 },
    { "supe", 8839 },
    { "szlig", 223 },
    { "tau", 964 },
    { "tcaron", 357 },
    { "tcedil", 355 },
    { "there4", 8756 },
    { "theta", 952 },
    { "thetasym", 977 },
    { "thorn", 254 },
    { "thorn", 254 },
    { "tilde", 732 },
    { "times", 215 },
    { "trade", 8482 },
    { "uArr", 8657 },
    { "uacute", 250 },
    { "uarr", 8593 },
    { "ucirc", 251 },
    { "udblac", 369 },
    { "ugrave", 249 },
    { "uml", 168 },
    { "upsih", 978 },
    { "upsilon", 965 },
    { "uring", 367 },
    { "uuml", 252 },
    { "weierp", 8472 },
    { "xi", 958 },
    { "yacute", 253 },
    { "yen", 165 },
    { "yuml", 255 },
    { "zacute", 378 },
    { "zcaron", 382 },
    { "zdot", 380 },
    { "zeta", 950 }
};



/* ------------------------------------------------------------------------ */
/*
 *   Write out a line.  Text we receive is in the UTF-8 character set.
 */
void CVmFormatter::write_text(VMG_ const wchar_t *txt, vm_nl_type nl)
{
    /* 
     *   Check the "script quiet" mode - this indicates that we're reading
     *   a script and not echoing output to the display.  If this mode is
     *   on, and we're writing to the display, suppress this write.  If
     *   the mode is off, or we're writing to a non-display stream (such
     *   as a log file stream), show the output as normal.  
     */
    if (!console_->is_quiet_script() || !is_disp_stream_)
    {
        char local_buf[128];
        char *dst;
        size_t rem;
        
        /*
         *   Check to see if we've reached the end of the screen, and if
         *   so run the MORE prompt.  Note that we don't make this check
         *   at all if USE_MORE is undefined, since this means that the OS
         *   layer code is taking responsibility for pagination issues.
         *   We also don't display a MORE prompt when reading from a
         *   script file.
         *   
         *   Note that we suppress the MORE prompt if the newline mode is
         *   VM_NL_NONE, since this is used to flush a partial line of
         *   text without starting a new line (for example, when
         *   displaying a prompt where the input will appear on the same
         *   line following the prompt).
         *   
         *   Skip the MORE prompt if this stream doesn't use it.  
         */
        if (use_more_mode()
            && !console_->is_reading_script()
            && console_->is_more_mode()
            && nl != VM_NL_NONE && nl != VM_NL_NONE_INTERNAL
            && linecnt_++ >= console_->get_page_length())
        {
            /* display the MORE prompt */
            console_->show_more_prompt(vmg0_);
        }

        /* convert and display the text */
        for (dst = local_buf, rem = sizeof(local_buf) - 1 ; *txt != '\0' ; )
        {
            size_t cur;
            size_t old_rem;
            
            /* try storing another character */
            old_rem = rem;
            cur = G_cmap_to_ui->map(*txt, &dst, &rem);

            /* if that failed, flush the buffer and try again */
            if (cur > old_rem)
            {
                /* null-terminate the buffer */
                *dst = '\0';
                
                /* display the text */
                disp_text_local_cs(local_buf);

                /* reset to the start of the local output buffer */
                dst = local_buf;
                rem = sizeof(local_buf) - 1;
            }
            else
            {
                /* we've now consumed this character of input */
                ++txt;
            }
        }

        /* if we have a partially-filled buffer, display it */
        if (dst > local_buf)
        {
            /* null-terminate and display the buffer */
            *dst = '\0';
            disp_text_local_cs(local_buf);
        }

        /* write the appropriate type of line termination */
        switch(nl)
        {
        case VM_NL_NONE:
        case VM_NL_NONE_INTERNAL:
            /* no line termination is needed */
            break;

        case VM_NL_NEWLINE:
            /* write a newline */
            disp_text_local_cs("\n");
            break;

        case VM_NL_OSNEWLINE:
            /* 
             *   the OS will provide a newline, but add a space to make it
             *   explicit that we can break the line here 
             */
            disp_text_local_cs(" ");
            break;
        }
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Flush the current line to the display, using the given type of line
 *   termination.
 *   
 *   VM_NL_NONE: flush the current line but do not start a new line; more
 *   text will follow on the current line.  This is used, for example, to
 *   flush text after displaying a prompt and before waiting for user
 *   input.
 *   
 *   VM_NL_NONE_INTERNAL: same as VM_NL_NONE, but doesn't flush at the OS
 *   level.  This is used when we're only flushing our buffers in order to
 *   clear out space internally, not because we want the underlying OS
 *   renderer to display things immediately.  This distinction is
 *   important in HTML mode, since it ensures that the HTML parser only
 *   sees well-formed strings when flushing.
 *   
 *   VM_NL_NEWLINE: flush the line and start a new line by writing out a
 *   newline character.
 *   
 *   VM_NL_OSNEWLINE: flush the line as though starting a new line, but
 *   don't add an actual newline character to the output, since the
 *   underlying OS display code will handle this.  Instead, add a space
 *   after the line to indicate to the OS code that a line break is
 *   possible there.  (This differs from VM_NL_NONE in that VM_NL_NONE
 *   doesn't add anything at all after the line.)  
 */

/* flush a given output stream */
void CVmFormatter::flush(VMG_ vm_nl_type nl)
{
    int i;
    vm_nl_type write_nl;

    /* null-terminate the current output line buffer */
    linebuf_[linepos_] = '\0';

    /* note the position of the last character to display */
    i = linepos_ - 1;

    /* if we're adding anything, remove trailing spaces */
    if (nl != VM_NL_NONE && nl != VM_NL_NONE_INTERNAL)
    {
        /* look for last non-space character */
        for ( ; i >= 0 && linebuf_[i] == ' ' ; --i) ;
    }

    /* null-terminate the buffer at the current position */
    linebuf_[++i] = '\0';

    /* check the newline mode */
    switch(nl)
    {
    case VM_NL_NONE:
    case VM_NL_NONE_INTERNAL:
        /* no newline - just flush out what we have */
        write_nl = VM_NL_NONE;
        break;

    case VM_NL_NEWLINE:
        /* 
         *   We're adding a newline.  We want to suppress redundant
         *   newlines -- we reduce any run of consecutive vertical
         *   whitespace to a single newline.  So, if we have anything in
         *   this line, or we didn't already just write a newline, write
         *   out a newline now; otherwise, write nothing.  
         */
        if (linecol_ != 0 || !just_did_nl_)
        {
            /* add the newline */
            write_nl = VM_NL_NEWLINE;
        }
        else
        {
            /* 
             *   Don't write out a newline after all - the line buffer is
             *   empty, and we just wrote a newline, so this is a
             *   redundant newline that we wish to suppress (so that we
             *   collapse a run of vertical whitespace down to a single
             *   newline).  
             */
            write_nl = VM_NL_NONE;
        }
        break;

    case VM_NL_OSNEWLINE:
        /* 
         *   we're going to depend on the underlying OS output layer to do
         *   line breaking, so we won't add a newline, but we will add a
         *   space, so that the underlying OS layer knows we have a word
         *   break here 
         */
        write_nl = VM_NL_OSNEWLINE;
        break;
    }

    /* 
     *   display the line, as long as we have something buffered to
     *   display; even if we don't, display it if our column is non-zero
     *   and we didn't just do a newline, since this must mean that we've
     *   flushed a partial line and are just now doing the newline 
     */
    if (linebuf_[preview_] != '\0'
        || (linecol_ != 0 && !just_did_nl_))
    {
        /* write it out */
        write_text(vmg_ linebuf_ + preview_, write_nl);
    }

    /* generate an HTML line break if necessary */
    if (nl == VM_NL_NEWLINE && html_mode_ && html_target_)
        write_text(vmg_ L"<BR HEIGHT=0>", VM_NL_NONE);

    /* check the line ending */
    if (nl == VM_NL_NONE)
    {
        /* we're not displaying a newline, so flush what we have */
        os_flush();
    }
    else
    {
        /* we displayed a newline, so reset the column position */
        linecol_ = 0;
    }

    /* reset the line output buffer position */
    linepos_ = preview_ = 0;

    /* 
     *   If we just output a newline, note it.  If we didn't just output a
     *   newline, but we did write out anything else, note that we're no
     *   longer at the start of a line on the underlying output device.  
     */
    if (nl == VM_NL_NEWLINE)
        just_did_nl_ = TRUE;
    else if (linebuf_[preview_] != '\0')
        just_did_nl_ = FALSE;
}


/* ------------------------------------------------------------------------ */
/*
 *   Display a blank line to the stream
 */
void CVmFormatter::write_blank_line(VMG0_)
{
    /* flush the stream */
    flush(vmg_ VM_NL_NEWLINE);

    /* if generating for an HTML display target, add an HTML line break */
    if (html_mode_ && html_target_)
        write_text(vmg_ L"<BR>", VM_NL_NONE);

    /* write out a blank line */
    write_text(vmg_ L"", VM_NL_NEWLINE);
}

/* ------------------------------------------------------------------------ */
/*
 *   Generate a tab for a "\t" sequence in the game text.
 *   
 *   Standard (non-HTML) version: we'll generate enough spaces to take us
 *   to the next tab stop.
 *   
 *   HTML version: if we're in native HTML mode, we'll just generate a
 *   <TAB MULTIPLE=4>; if we're not in HTML mode, we'll generate a hard
 *   tab character, which the HTML formatter will interpret as a <TAB
 *   MULTIPLE=4>.  
 */
void CVmFormatter::write_tab(VMG0_)
{
    /* check to see what the underlying system is expecting */
    if (html_target_)
    {
        /* the underlying system is HTML - check for HTML mode */
        if (html_mode_)
        {
            /* we're in HTML mode, so use the HTML <TAB> tag */
            buffer_string(vmg_ "<TAB MULTIPLE=4>");
        }
        else
        {
            /* we're not in HTML mode, so generate a hard tab character */
            buffer_char(vmg_ QTAB);
        }
    }
    else
    {
        /* 
         *   We're not in HTML mode - expand the tab with spaces.  Keep
         *   going until we reach the next tab stop; our tab stops are
         *   every four characters, hence there's a tab stop at each
         *   multiple of four.  
         */
        do
        {
            /* add another space */
            linebuf_[linepos_++] = ' ';

            /* advance the column counter */
            ++linecol_;
        } while (((linecol_ + 1) & 3) != 0 && linecol_ < OS_MAXWIDTH);
    }
}


/* ------------------------------------------------------------------------ */
/*
 *   Flush a line 
 */
void CVmFormatter::flush_line(VMG_ int padding)
{
    /* 
     *   check to see if we're using the underlying display layer's line
     *   wrapping 
     */
    if (os_line_wrap_)
    {
        /*
         *   In the HTML version, we don't need the normal *MORE*
         *   processing, since the HTML layer will handle that.
         *   Furthermore, we don't need to provide actual newline breaks
         *   -- that happens after the HTML is parsed, so we don't have
         *   enough information here to figure out actual line breaks.
         *   So, we'll just flush out our buffer whenever it fills up, and
         *   suppress newlines.
         *   
         *   Similarly, if we have OS-level line wrapping, don't try to
         *   figure out where the line breaks go -- just flush our buffer
         *   without a trailing newline whenever the buffer is full, and
         *   let the OS layer worry about formatting lines and paragraphs.
         *   
         *   If we're using padding, use newline mode VM_NL_OSNEWLINE.  If
         *   we don't want padding (which is the case if we completely
         *   fill up the buffer without finding any word breaks), write
         *   out in mode VM_NL_NONE, which just flushes the buffer exactly
         *   like it is.  
         */
        flush(vmg_ padding ? VM_NL_OSNEWLINE : VM_NL_NONE_INTERNAL);
    }
    else
    {
        /*
         *   Normal mode - we process the *MORE* prompt ourselves, and we
         *   are responsible for figuring out where the actual line breaks
         *   go.  Use flush() to generate an actual newline whenever we
         *   flush out our buffer.  
         */
        flush(vmg_ VM_NL_NEWLINE);
    }
}


/* ------------------------------------------------------------------------ */
/*
 *   Write a character to an output stream.  The character is provided to
 *   us as a wide Unicode character.  
 */
void CVmFormatter::buffer_char(VMG_ wchar_t c)
{
    wchar_t brkchar;
    int  i;
    int  qspace;
    
    /* check for the special quoted space character */
    if (c == QSPACE)
    {
        /* it's a quoted space - note it and convert it to a regular space */
        qspace = TRUE;
        c = ' ';
    }
    else if (c == QTAB)
    {
        /* 
         *   it's a hard tab - put a hard tab character in the buffer for
         *   passing to the underlying renderer (the only time we use this
         *   character is for rendering a tab to an underlying HTML
         *   processor; in any other case we will already have expanded
         *   any tabs into runs of spaces) 
         */
        c = '\t';
        qspace = FALSE;
    }
    else
    {
        /* 
         *   Translate any whitespace character to a regular space
         *   character.  Note that, once this is done, we don't need to
         *   worry about calling t3_is_space() any more - we can use the
         *   faster is_space() from now on, because we know we have a
         *   regular space (or possibly a regular tab, from above). 
         */
        if (t3_is_space(c))
            c = ' ';
        
        /* 
         *   it's not a quoted space - treat it as normal space unless
         *   we're in obey-whitespace mode, in which case treat it as a
         *   non-breaking space 
         */
        qspace = obey_whitespace_;
    }

    /* check for the caps/nocaps flags */
    if ((capsflag_ || allcapsflag_) && t3_is_alpha(c))
    {
        /* capsflag is set, so capitalize this character */
        c = t3_to_upper(c);

        /* okay, we've capitalized something; clear flag */
        capsflag_ = FALSE;
    }
    else if (nocapsflag_ && t3_is_alpha(c))
    {
        /* nocapsflag is set, so minisculize this character */
        c = t3_to_lower(c);

        /* clear the flag now that we've done the job */
        nocapsflag_ = FALSE;
    }

    /* add the character to out output buffer, flushing as needed */
    if (linecol_ + 1 < console_->get_line_width())
    {
        /* 
         *   there's room for this character, so add it to the buffer 
         */
        
        /* 
         *   Ignore non-quoted space at start of line.  (Note that we use
         *   is_space() rather than t3_is_space() - we know we've already
         *   translated this character to a regular space or a regular
         *   tab, so we don't need to make the more extensive and
         *   expensive check that t3_is_space() makes; is_space() is
         *   perfectly adequate for our needs here.) 
         */
        if (is_space(c) && linecol_ == 0 && !qspace)
            return;

        /* is this a non-quoted space not at the start of the line? */
        if (is_space(c) && linecol_ != 0 && !qspace)
        {
            int pos1;
            wchar_t prv;

            /* check the previous character */
            pos1 = linepos_ - 1;
            prv = linebuf_[pos1];

            /* ignore repeated spaces - collapse into a single space */
            if (is_space(prv))
                return;

            /*
             *   Certain punctuation requires a double space: a period, a
             *   question mark, an exclamation mark, or a colon; or any of
             *   these characters followed by any number of single and/or
             *   double quotes.  First, scan back to before any quotes, if
             *   are on one now, then check the preceding character; if
             *   it's one of the punctuation marks requiring a double
             *   space, add this space a second time.  (In addition to
             *   scanning back past quotes, scan past parentheses,
             *   brackets, and braces.)  Don't double the spacing if we're
             *   not in the normal doublespace mode; some people may
             *   prefer single spacing after punctuation, so we make this
             *   a run-time option.  
             */
            if (console_->get_doublespace())
            {
                /* 
                 *   Find the previous relevant punctuation character.
                 *   Skip certain punctuation - in particular, skip
                 *   quotes, close parentheses, braces, and brackets,
                 *   since we want the effect of any double-spacing
                 *   punctuation to transcend these grouping marks. 
                 */
                while (pos1 != 0 &&
                       (prv == '"' || prv == '\'' || prv == ')' || prv == ']'
                        || prv == '}' || prv == get_hilite_char(1)
                        || prv == get_hilite_char(2)))
                {
                    /* go back one position */
                    prv = linebuf_[--pos1];
                }

                /* check for double-spacing punctuation */
                if (prv == '.' || prv == '?' || prv == '!' || prv == ':')
                {
                    /* a double-space is required after this character */
                    linebuf_[linepos_++] = c;
                    ++linecol_;
                }
            }
        }

        /* add this character to the buffer */
        linebuf_[linepos_++] = c;

        /* advance the output column position */
        ++linecol_;
        return;
    }

    /*
     *   The line would overflow if this character were added.  Find the
     *   most recent word break, and output the line up to the previous
     *   word.  Note that if we're trying to output a space, we'll just
     *   add it to the line buffer.  If the last character of the line
     *   buffer is already a space, we won't do anything right now.  
     */
    if (is_space(c))
    {
        /* this is a space, so we're at a word break */
        if (!is_space(linebuf_[linepos_ - 1]))
            linebuf_[linepos_++] = ' ';
        return;
    }
    
    /*
     *   Find the most recent word break: look for a space or a dash.
     *   Note that when "more" mode isn't active (which means that line
     *   breaks will be handled by the OS-level display code rather than
     *   here), we will NOT break on dashes.  Note that if we're about to
     *   write a hyphen, skip all contiguous hyphens, because we want to
     *   keep them together as a single punctuation mark; then keep going
     *   in the normal manner, which will keep the hyphens plus the word
     *   they're attached to together as a single unit.  If spaces precede
     *   the sequence of hyphens, include the prior word as well.  
     */
    i = linepos_ - 1;
    if (c == '-')
    {
        /* skip any contiguous hyphens at the end of the line */
        for ( ; i >= 0 && linebuf_[i] == '-' ; --i) ;
        
        /* skip any spaces preceding the sequence of hyphens */
        for ( ; i >= 0 && is_space(linebuf_[i]) ; --i) ;
    }

    /* 
     *   Now find the preceding space.  If we're doing our own wrapping
     *   (i.e., we're not using OS line wrapping), then look for the
     *   nearest hyphen as well. 
     */
    for ( ; i >= 0 && !is_space(linebuf_[i])
          && !(!os_line_wrap_ && linebuf_[i] == '-') ; --i) ;

    /* check to see if we found a good place to break */
    if (i < 0)
    {
        /* 
         *   we didn't find any good place to break - flush the entire
         *   line as-is, breaking arbitrarily in the middle of a word 
         */
        flush_line(vmg_ FALSE);

        /* 
         *   we've completely cleared out the line buffer, so reset all of
         *   the line buffer counters 
         */
        linepos_ = 0;
        linecol_ = 0;
        linebuf_[0] = '\0';
    }
    else
    {
        wchar_t tmpbuf[OS_MAXWIDTH];

        /* remember word-break character */        
        brkchar = linebuf_[i];

        /* null-terminate the line buffer */        
        linebuf_[linepos_] = '\0';

        /* the next line starts after the break - save a copy */
        wcscpy(tmpbuf, &linebuf_[i+1]);

        /* 
         *   terminate the buffer at the space or after the hyphen,
         *   depending on where we broke 
         */
        if (is_space(brkchar))
            linebuf_[i] = '\0';
        else
            linebuf_[i+1] = '\0';

        /* write out everything up to the word break */
        flush_line(vmg_ TRUE);

        /* move next line into line buffer */
        wcscpy(linebuf_, tmpbuf);
        linepos_ = wcslen(linebuf_);

        /* 
         *   figure what column we're now in - count all of the printable
         *   characters in the new line 
         */
        for (linecol_ = 0, i = 0 ; i < linepos_ ; ++i)
        {
            /* if it's printable, count it */
            if (linebuf_[i] >= 26)
                ++linecol_;
        }
    }
    
    /* add the new character to buffer */
    linebuf_[linepos_++] = c;

    /* advance the column counter */
    ++linecol_;
}

/* ------------------------------------------------------------------------ */
/* 
 *   write out a UTF-8 string
 */
void CVmFormatter::buffer_string(VMG_ const char *txt)
{
    /* write out each character in the string */
    for ( ; utf8_ptr::s_getch(txt) != 0 ; txt += utf8_ptr::s_charsize(*txt))
        buffer_char(vmg_ utf8_ptr::s_getch(txt));
}

/*
 *   write out a wide unicode string 
 */
void CVmFormatter::buffer_wstring(VMG_ const wchar_t *txt)
{
    /* write out each wide character */
    for ( ; *txt != '\0' ; ++txt)
        buffer_char(vmg_ *txt);
}


/* ------------------------------------------------------------------------ */
/*
 *   Get the next wide unicode character in a UTF8-encoded string, and
 *   update the string pointer and remaining length.  Returns zero if no
 *   more characters are available in the string.  
 */
wchar_t CVmFormatter::next_wchar(const char **s, size_t *len)
{
    wchar_t ret;
    size_t charsize;

    /* if there's nothing left, return a null terminator */
    if (*len == 0)
        return 0;

    /* get this character */
    ret = utf8_ptr::s_getch(*s);

    /* advance the string pointer and length counter */
    charsize = utf8_ptr::s_charsize(**s);
    *len -= charsize;
    *s += charsize;

    /* return the result */
    return ret;
}

/* ------------------------------------------------------------------------ */
/*
 *   Display a string of a given length.  The text is encoded as UTF-8
 *   characters.  
 */
int CVmFormatter::format_text(VMG_ const char *s, size_t slen)
{
    wchar_t c;
    int done = FALSE;

    /* get the first character */
    c = next_wchar(&s, &slen);

    /* if we have anything to show, show it */
    while (c != '\0')
    {
        /*
         *   If we're parsing HTML here, and we're inside a tag, skip
         *   characters until we reach the end of the tag.  
         */
        if (html_mode_flag_ != HTML_MODE_NORMAL)
        {
            switch(html_mode_flag_)
            {
            case HTML_MODE_TAG:
                /* 
                 *   keep skipping up to the closing '>', but note when we
                 *   enter any quoted section 
                 */
                switch(c)
                {
                case '>':
                    /* we've reached the end of the tag */
                    html_mode_flag_ = HTML_MODE_NORMAL;

                    /* if we have a deferred <BR>, process it now */
                    switch(html_defer_br_)
                    {
                    case HTML_DEFER_BR_NONE:
                        /* no deferred <BR> */
                        break;

                    case HTML_DEFER_BR_FLUSH:
                        flush(vmg_ VM_NL_NEWLINE);
                        break;

                    case HTML_DEFER_BR_BLANK:
                        write_blank_line(vmg0_);
                        break;
                    }

                    /* no more deferred <BR> pending */
                    html_defer_br_ = HTML_DEFER_BR_NONE;

                    /* no more ALT attribute allowed */
                    html_allow_alt_ = FALSE;
                    break;

                case '"':
                    /* enter a double-quoted string */
                    html_mode_flag_ = HTML_MODE_DQUOTE;
                    break;

                case '\'':
                    /* enter a single-quoted string */
                    html_mode_flag_ = HTML_MODE_SQUOTE;
                    break;

                default:
                    /* if it's alphabetic, note the attribute name */
                    if (t3_is_alpha(c))
                    {
                        const size_t MAX_ATTR_NAME = 128;
                        const size_t MAX_ATTR_VAL = 256;
                        wchar_t attrname[MAX_ATTR_NAME];
                        wchar_t attrval[MAX_ATTR_VAL];
                        wchar_t *dst;

                        /* gather up the attribute name */
                        for (dst = attrname ;
                             dst + 1 < attrname + MAX_ATTR_NAME ; )
                        {
                            /* store this character */
                            *dst++ = c;

                            /* get the next character */
                            c = next_wchar(&s, &slen);

                            /* if it's not alphanumeric, stop scanning */
                            if (!t3_is_alpha(c) && !t3_is_digit(c))
                                break;
                        }

                        /* null-terminate the result */
                        *dst++ = '\0';

                        /* gather the value if present */
                        if (c == '=')
                        {
                            wchar_t qu;
                            
                            /* skip the '=' */
                            c = next_wchar(&s, &slen);

                            /* if we have a quote, so note */
                            if (c == '"' || c == '\'')
                            {
                                /* remember the quote */
                                qu = c;

                                /* skip it */
                                c = next_wchar(&s, &slen);
                            }
                            else
                            {
                                /* no quote */
                                qu = '\0';
                            }

                            /* read the value */
                            for (dst = attrval ;
                                 dst + 1 < attrval + MAX_ATTR_VAL ; )
                            {
                                /* store this character */
                                *dst++ = c;

                                /* read the next one */
                                c = next_wchar(&s, &slen);
                                if (c == '\0')
                                {
                                    /* 
                                     *   we've reached the end of the
                                     *   string, and we're still inside
                                     *   this attribute - abandon the
                                     *   attribute but note that we're
                                     *   inside a quoted string if
                                     *   necessary 
                                     */
                                    if (qu == '"')
                                        html_mode_flag_ = HTML_MODE_DQUOTE;
                                    else if (qu == '\'')
                                        html_mode_flag_ = HTML_MODE_SQUOTE;
                                    else
                                        html_mode_flag_ = HTML_MODE_TAG;

                                    /* stop scanning the string */
                                    break;
                                }

                                /* 
                                 *   if we're looking for a quote, check
                                 *   for the closing quote; otherwise,
                                 *   check for alphanumerics 
                                 */
                                if (qu != 0)
                                {
                                    /* if this is our quote, stop scanning */
                                    if (c == qu)
                                        break;
                                }
                                else
                                {
                                    /* if it's non-alphanumeric, we're done */
                                    if (!t3_is_alpha(c) && !t3_is_digit(c))
                                        break;
                                }
                            }

                            /* skip the closing quote, if necessary */
                            if (qu != 0 && c == qu)
                                c = next_wchar(&s, &slen);
                            
                            /* null-terminate the value string */
                            *dst = '\0';
                        }
                        else
                        {
                            /* no value */
                            attrval[0] = '\0';
                        }
                        
                        /* 
                         *   see if we recognize it, and it's meaningful
                         *   in the context of the current tag 
                         */
                        if (w_stricmp(attrname, L"height") == 0
                            && html_defer_br_ != HTML_DEFER_BR_NONE)
                        {
                            int ht;
                                
                            /*
                             *   If the height is zero, always treat this
                             *   as a non-blanking flush.  If it's one,
                             *   treat it as we originally planned to.  If
                             *   it's greater than one, add n blank lines.
                             */
                            ht = wtoi(attrval);
                            if (ht == 0)
                            {
                                /* always use non-blanking flush */
                                html_defer_br_ = HTML_DEFER_BR_FLUSH;
                            }
                            else if (ht == 1)
                            {
                                /* keep original setting */
                            }
                            else
                            {
                                /* 
                                 *   write out the desired number of blank
                                 *   lines 
                                 */
                                for ( ; ht > 0 ; --ht)
                                    write_blank_line(vmg0_);
                            }
                        }
                        else if (w_stricmp(attrname, L"alt") == 0
                                 && html_allow_alt_)
                        {
                            /* write out the ALT string */
                            buffer_wstring(vmg_ attrval);
                        }

                        /* 
                         *   since we already read the next character,
                         *   simply loop back immediately 
                         */
                        continue;
                    }
                    break;
                }
                break;

            case HTML_MODE_DQUOTE:
                /* if we've reached the closing quote, return to tag state */
                if (c == '"')
                    html_mode_flag_ = HTML_MODE_TAG;
                break;

            case HTML_MODE_SQUOTE:
                /* if we've reached the closing quote, return to tag state */
                if (c == '\'')
                    html_mode_flag_ = HTML_MODE_TAG;
                break;
            }

            /* 
             *   move on to the next character, and start over with the
             *   new character 
             */
            c = next_wchar(&s, &slen);
            continue;
        }

        /*
         *   If we're in a title, and this isn't the start of a new tag,
         *   skip the character - we suppress all regular text output
         *   inside a <TITLE> ... </TITLE> sequence. 
         */
        if (html_in_ignore_ && c != '<')
        {
            /* 
             *   if we're gathering a title, and there's room in the title
             *   buffer for more (always leaving room for a null
             *   terminator), add this to the title buffer 
             */
            if (html_in_title_)
            {
                size_t rem, orig_rem;

                /* 
                 *   figure the remaining title buffer space, leaving room
                 *   for a terminating null byte 
                 */
                orig_rem = rem = HTML_TITLE_BUF_SIZE - 1;
                
                /* 
                 *   map this character into the local character set and
                 *   add it to the buffer 
                 */
                if (G_cmap_to_ui->map(c, &html_title_ptr_, &rem) > orig_rem)
                {
                    /* 
                     *   it didn't fit - add null termination and close
                     *   out the buffer (we might have to add more than
                     *   one byte, because a single character might take
                     *   up multiple bytes in the local character set; to
                     *   ensure that we don't skip a wide character then
                     *   later add a narrower character, close out the
                     *   entire buffer when first we encounter a character
                     *   that doesn't fit)
                     */
                    while (html_title_ptr_
                           < html_title_buf_ + HTML_TITLE_BUF_SIZE)
                        *html_title_ptr_++ = '\0';
                }
            }

            /* get the next character */
            c = next_wchar(&s, &slen);

            /* don't display anything in an ignore section */
            continue;
        }

        /* check for special characters */
        switch(c)
        {
        case 0x0013:
            /* 
             *   \H+ - turn on HTML mode in the renderer; if we have an
             *   HTML target, notify it 
             */
            if (html_target_)
            {
                /* flush the underlying stream */
                flush(vmg_ VM_NL_NONE);
                
                /* tell the OS layer to switch to HTML mode */
                start_html();
            }
            
            /* switch to HTML mode internally */
            html_mode_ = TRUE;
            break;

        case 0x0014:
            /* 
             *   \H- - turn off HTML mode in the renderer; if we have an
             *   HTML target, notify it 
             */
            if (html_target_)
            {
                /* flush its stream */
                flush(vmg_ VM_NL_NONE);
                
                /* tell the OS layer to switch to normal mode */
                end_html();
            }

            /* switch to normal mode internally */
            html_mode_ = FALSE;
            break;
            
        case 10:
            /* newline */
            flush(vmg_ VM_NL_NEWLINE);
            break;
                    
        case 9:
            /* tab */
            write_tab(vmg0_);
            break;

        case 0x000B:
            /* \b - blank line */
            write_blank_line(vmg0_);
            break;
                    
        case 0x0015:
            /* non-breaking space */
            if (html_target_ && html_mode_)
            {
                /* 
                 *   we're generating for an HTML target and we're in HTML
                 *   mode - generate the HTML non-breaking space 
                 */
                buffer_string(vmg_ "&nbsp;");
            }
            else
            {
                /* 
                 *   we're not in HTML mode - generate our internal quoted
                 *   space character 
                 */
                buffer_char(vmg_ QSPACE);
            }
            break;

        case 0x000F:
            /* capitalize next character */
            capsflag_ = TRUE;
            nocapsflag_ = FALSE;
            break;

        case 0x000E:
            nocapsflag_ = TRUE;
            capsflag_ = FALSE;
            break;

        case 0x0011:
            /* highlighting on - generate HTML if in the appropriate mode */
            if (html_mode_ && html_target_)
                buffer_string(vmg_ "<B>");

            /* generate the OS-level escape code */
            if (get_hilite_char(2))
                buffer_char(vmg_ get_hilite_char(2));
            break;

        case 0x0012:
            /* highlighting off - generate HTML if in the appropriate mode */
            if (html_mode_ && html_target_)
                buffer_string(vmg_ "</B>");

            /* generate the OS-level escape code */
            if (get_hilite_char(1))
                buffer_char(vmg_ get_hilite_char(1));
            break;

        case '<':
        case '&':
            /* HTML markup-start character */
            if (!html_target_ && html_mode_)
            {
                /*
                 *   We're in HTML mode, but the underlying target does
                 *   not accept HTML sequences.  It appears we're at the
                 *   start of an "&" entity or a tag sequence, so parse
                 *   it, remove it, and replace it (if possible) with a
                 *   text-only equivalent.  
                 */
                if (c == '<')
                {
                    const size_t MAX_TAG_SIZE = 50;
                    wchar_t tagbuf[MAX_TAG_SIZE];
                    wchar_t *dst;
                    int is_end_tag;
                    
                    /* skip the opening '<' */
                    c = next_wchar(&s, &slen);

                    /* note if this is a closing tag */
                    if (c == '/' || c == '\\')
                    {
                        /* it's an end tag - note it and skip the slash */
                        is_end_tag = TRUE;
                        c = next_wchar(&s, &slen);
                    }
                    else
                        is_end_tag = FALSE;
                
                    /* 
                     *   find the end of the tag name - the tag continues
                     *   to the next space, '>', or end of line 
                     */
                    for (dst = tagbuf ; c != '\0' && c != ' ' && c != '>' ;
                         c = next_wchar(&s, &slen))
                    {
                        /* add this to the tag buffer if it fits */
                        if (dst + 1 < tagbuf + MAX_TAG_SIZE)
                            *dst++ = c;
                    }
                    
                    /* null-terminate the tag name */
                    *dst = '\0';
                    
                    /*
                     *   Check to see if we recognize the tag.  We only
                     *   recognize a few simple tags that map easily to
                     *   character mode.  
                     */
                    if (w_stricmp(tagbuf, L"br") == 0)
                    {
                        /* 
                         *   line break - if there's anything buffered up,
                         *   just flush the current line, otherwise write
                         *   out a blank line 
                         */
                        if (html_in_ignore_)
                            /* suppress in ignore mode */;
                        else if (linepos_ != 0)
                            html_defer_br_ = HTML_DEFER_BR_FLUSH;
                        else
                            html_defer_br_ = HTML_DEFER_BR_BLANK;
                    }
                    else if (w_stricmp(tagbuf, L"b") == 0
                             || w_stricmp(tagbuf, L"em") == 0)
                    {
                        /* suppress in ignore mode */
                        if (!html_in_ignore_)
                        {
                            /* 
                             *   bold on/off - send out the appropriate
                             *   os-layer code 
                             */
                            if (!is_end_tag) 
                            {
                                if (get_hilite_char(2))
                                    buffer_char(vmg_ get_hilite_char(2));
                            }
                            else
                            {
                                if (get_hilite_char(1))
                                    buffer_char(vmg_ get_hilite_char(1));
                            }
                        }
                    }
                    else if (w_stricmp(tagbuf, L"p") == 0)
                    {
                        /* paragraph - send out a blank line */
                        if (!html_in_ignore_)
                            write_blank_line(vmg0_);
                    }
                    else if (w_stricmp(tagbuf, L"tab") == 0)
                    {
                        /* tab - send out a \t */
                        if (!html_in_ignore_)
                            write_tab(vmg0_);
                    }
                    else if (w_stricmp(tagbuf, L"img") == 0
                             || w_stricmp(tagbuf, L"sound") == 0)
                    {
                        /* IMG and SOUND - allow ALT attributes */
                        html_allow_alt_ = TRUE;
                    }
                    else if (w_stricmp(tagbuf, L"hr") == 0)
                    {
                        int rem;

                        if (!html_in_ignore_)
                        {
                            /* start a new line */
                            flush(vmg_ VM_NL_NEWLINE);
                            
                            /* write out underscores to the display width */
                            for (rem = console_->get_line_width() - 1 ;
                                 rem > 0 ; )
                            {
                                const size_t DASHBUFLEN = 100;
                                wchar_t dashbuf[DASHBUFLEN];
                                wchar_t *p;
                                int cur;
                                
                                /* do as much as we can on this pass */
                                cur = rem;
                                if (cur > DASHBUFLEN - 1)
                                    cur = DASHBUFLEN - 1;
                                
                                /* deduct this from the total */
                                rem -= cur;
                                
                                /* fill the buffer with dashes */
                                for (p = dashbuf ; cur != 0 ; --cur)
                                    *p++ = '-';
                                *p = '\0';
                                
                                /* write it out */
                                buffer_wstring(vmg_ dashbuf);
                            }
                            
                            /* put a blank line after the underscores */
                            write_blank_line(vmg0_);
                        }
                    }
                    else if (w_stricmp(tagbuf, L"q") == 0)
                    {
                        wchar_t htmlchar;

                        if (!html_in_ignore_)
                        {
                            /* if it's an open quote, increment the level */
                            if (!is_end_tag)
                                ++html_quote_level_;
                            
                            /* add the open quote */
                            htmlchar =
                                (!is_end_tag
                                 ? ((html_quote_level_ & 1) == 1
                                    ? 8220 : 8216)
                                 : ((html_quote_level_ & 1) == 1
                                    ? 8221 : 8217));
                            
                            /* write out the HTML quote character */
                            buffer_char(vmg_ htmlchar);
                            
                            /* if it's a close quote, decrement the level */
                            if (is_end_tag)
                                --html_quote_level_;
                        }
                    }
                    else if (w_stricmp(tagbuf, L"title") == 0)
                    {
                        /* 
                         *   Turn ignore mode on or off as appropriate,
                         *   and turn on or off title mode as well.  
                         */
                        if (is_end_tag)
                        {
                            /* 
                             *   note that we're leaving an ignore section
                             *   and a title section 
                             */
                            --html_in_ignore_;
                            --html_in_title_;

                            /*  
                             *   if we're no longer in a title, call the
                             *   OS layer to tell it the title string, in
                             *   case it wants to change the window title
                             *   or otherwise make use of the title 
                             */
                            if (html_in_title_ == 0)
                            {
                                /* null-terminate the title string */
                                *html_title_ptr_ = '\0';
                                
                                /* tell the OS about the title */
                                os_set_title(html_title_buf_);
                            }
                        }
                        else
                        {
                            /* 
                             *   if we aren't already in a title, set up
                             *   to capture the title into the title
                             *   buffer 
                             */
                            if (!html_in_title_)
                                html_title_ptr_ = html_title_buf_;
                            
                            /* 
                             *   note that we're in a title and in an
                             *   ignore section, since nothing within gets
                             *   displayed 
                             */
                            ++html_in_ignore_;
                            ++html_in_title_;
                        }
                    }
                    else if (w_stricmp(tagbuf, L"aboutbox") == 0)
                    {
                        /* turn ignore mode on or off as appropriate */
                        if (is_end_tag)
                            --html_in_ignore_;
                        else
                            ++html_in_ignore_;
                    }
                    
                    /* suppress everything up to the next '>' */
                    html_mode_flag_ = HTML_MODE_TAG;
                    
                    /* 
                     *   continue with the current character; since we're in
                     *   html tag mode, we'll skip everything until we get to
                     *   the closing '>' 
                     */
                    continue;
                }
                else if (c == '&')
                {
                    const size_t MAXAMPLEN = 10;
                    char ampbuf[MAXAMPLEN];
                    char *dst;
                    const char *orig_s;
                    size_t orig_slen;
                    const struct amp_tbl_t *ampptr;
                    size_t lo, hi, cur;
                    
                    /* 
                     *   remember where the part after the '&' begins, so
                     *   we can come back here later if necessary 
                     */
                    orig_s = s;
                    orig_slen = slen;
                    
                    /* get the character after the ampersand */
                    c = next_wchar(&s, &slen);
                    
                    /* if it's numeric, parse the number */
                    if (c == '#')
                    {
                        uint val;
                        
                        /* skip the '#' */
                        c = next_wchar(&s, &slen);
                        
                        /* check for hex */
                        if (c == 'x' || c == 'X')
                        {
                            /* skip the 'x' */
                            c = next_wchar(&s, &slen);
                            
                            /* read the hex number */
                            for (val = 0 ; is_xdigit(c) ;
                                 c = next_wchar(&s, &slen))
                            {
                                /* accumulate this digit into the value */
                                val *= 16;
                                if (is_digit(c))
                                    val += c - '0';
                                else if (c >= 'a' && c <= 'f')
                                    val += c - 'a' + 10;
                                else
                                    val += c - 'A' + 10;
                            }
                        }
                        else
                        {
                            /* read the number */
                            for (val = 0 ; is_digit(c) ;
                                 c = next_wchar(&s, &slen))
                            {
                                /* accumulate this digit into the value */
                                val *= 10;
                                val += c - '0';
                            }
                        }

                        /* if we found a ';' at the end, skip it */
                        if (c == ';')
                            c = next_wchar(&s, &slen);

                        /* translate and write the character */
                        buffer_char(vmg_ (wchar_t)val);

                        /* we're done with this character */
                        continue;
                    }

                    /*
                     *   Parse the sequence after the '&'.  Parse up to
                     *   the closing semicolon, or any non-alphanumeric,
                     *   or until we fill up the buffer.  
                     */
                    for (dst = ampbuf ;
                         c != '\0' && (is_digit(c) || is_alpha(c))
                             && dst < ampbuf + MAXAMPLEN - 1 ;
                         c = next_wchar(&s, &slen))
                    {
                        /* copy this character to the name buffer */
                        *dst++ = (char)c;
                    }

                    /* null-terminate the name */
                    *dst = '\0';
                    
                    /* do a binary search for the name */
                    lo = 0;
                    hi = sizeof(amp_tbl)/sizeof(amp_tbl[0]) - 1;
                    for (;;)
                    {
                        int diff;
                        
                        /* if we've converged, look no further */
                        if (lo > hi
                            || lo >= sizeof(amp_tbl)/sizeof(amp_tbl[0]))
                        {
                            ampptr = 0;
                            break;
                        }
                        
                        /* split the difference */
                        cur = lo + (hi - lo)/2;
                        ampptr = &amp_tbl[cur];
                        
                        /* see where we are relative to the target item */
                        diff = strcmp(ampptr->cname, ampbuf);
                        if (diff == 0)
                        {
                            /* this is it */
                            break;
                        }
                        else if (diff > 0)
                        {
                            /* this one is too high - check the lower half */
                            hi = (cur == hi ? hi - 1 : cur);
                        }
                        else
                        {
                            /* this one is too low - check the upper half */
                            lo = (cur == lo ? lo + 1 : cur);
                        }
                    }
                    
                    /* skip to the appropriate next character */
                    if (c == ';')
                    {
                        /* name ended with semicolon - skip the semicolon */
                        c = next_wchar(&s, &slen);
                    }
                    else if (ampptr != 0)
                    {
                        int skipcnt;
                        
                        /* found the name - skip its exact length */
                        skipcnt = strlen(ampptr->cname);
                        for (s = orig_s, slen = orig_slen ; skipcnt != 0 ;
                             c = next_wchar(&s, &slen), --skipcnt) ;
                    }

                    /* if we found the entry, write out the character */
                    if (ampptr != 0)
                    {
                        /* we found it - write it out */
                        buffer_char(vmg_ ampptr->html_cval);
                    }
                    else
                    {
                        /* 
                         *   we didn't find it - ignore the entire
                         *   sequence, and just write it out verbatim;
                         *   start by writing out the ampersand 
                         */
                        buffer_char(vmg_ '&');

                        /* 
                         *   now go back and scan from the next character
                         *   after the ampersand 
                         */
                        s = orig_s;
                        slen = orig_slen;
                        c = next_wchar(&s, &slen);
                    }

                    /* proceed with the next character */
                    continue;
                }
            }
            else
            {
                /* 
                 *   we're either not in HTML mode, or the underlying OS
                 *   renderer interprets HTML mark-up sequences -- in
                 *   either case, we don't need to perform any
                 *   interpretation; simply pass through the character as
                 *   though it were any other 
                 */
                goto normal_char;
            }
            break;

        default:
        normal_char:
            /* normal character - write it out */
            buffer_char(vmg_ c);
        }

        /* move on to the next character, unless we're finished */
        if (done)
            c = '\0';
        else
            c = next_wchar(&s, &slen);
    }

    /* success */
    return 0;
}


/* ------------------------------------------------------------------------ */
/*
 *   Initialize the display object 
 */
CVmConsole::CVmConsole()
{
    /* no script file yet */
    script_fp_ = 0;
    quiet_script_ = FALSE;

    /* no command log file yet */
    command_fp_ = 0;

    /* assume we'll double-space after each period */
    doublespace_ = TRUE;

    /* clear the debug flags */
    outwxflag_ = FALSE;

    /* create and initialize our display stream */
    disp_str_ = new CVmFormatterDisp(this);
    disp_str_->init();

    /* create and initialize our log stream */
    log_str_ = new CVmFormatterLog(this);
    log_str_->init();
}

/*
 *   Delete the display object 
 */
CVmConsole::~CVmConsole()
{
    /* close any active script file */
    close_script_file();

    /* close any active command log file */
    close_command_log();

    /* delete the output formatter streams */
    delete disp_str_;
    delete log_str_;
}

/* ------------------------------------------------------------------------ */
/*
 *   Display a string of a given byte length 
 */
int CVmConsole::format_text(VMG_ const char *p, size_t len)
{
    int ret;

    /* presume we'll return success */
    ret = 0;

    /* if the debugger is showing watchpoints, suppress all output */
    if (outwxflag_)
        return ret;

    /* display the string */
    disp_str_->format_text(vmg_ p, len);

    /* if there's a log file, write to the log file as well */
    log_str_->format_text(vmg_ p, len);

    /* return the result from displaying to the screen */
    return ret;
}

/* ------------------------------------------------------------------------ */
/*
 *   Display a blank line 
 */
void CVmConsole::write_blank_line(VMG0_)
{
    /* generate the newline to the standard display */
    disp_str_->write_blank_line(vmg0_);

    /* if we're logging, generate the newline to the log file as well */
    log_str_->write_blank_line(vmg0_);
}


/* ------------------------------------------------------------------------ */
/*
 *   outcaps() - sets an internal flag which makes the next letter output
 *   a capital, whether it came in that way or not.  Set the same state in
 *   both formatters (standard and log).  
 */
void CVmConsole::caps()
{
    disp_str_->caps();
    log_str_->caps();
}

/*
 *   outnocaps() - sets the next letter to a miniscule, whether it came in
 *   that way or not.  
 */
void CVmConsole::nocaps()
{
    disp_str_->nocaps();
    log_str_->nocaps();
}

/*
 *   obey_whitespace() - sets the obey-whitespace mode 
 */
int CVmConsole::set_obey_whitespace(int f)
{
    int ret;

    /* note the original display stream status */
    ret = disp_str_->get_obey_whitespace();

    /* set the stream status */
    disp_str_->set_obey_whitespace(f);
    log_str_->set_obey_whitespace(f);

    /* return the original status of the display stream */
    return ret;
}

/* ------------------------------------------------------------------------ */
/*
 *   Open a log file 
 */
int CVmConsole::open_log_file(const char *fname)
{
    /* 
     *   Tell the log stream to open the file.  Set the log file's HTML
     *   source mode flag to the same value as is currently being used in
     *   the main display stream, so that it will interpret source markups
     *   the same way that the display stream is going to.  
     */
    return log_str_->open_log_file(fname, disp_str_->is_html_mode());
}

/*
 *   Close the log file 
 */
int CVmConsole::close_log_file()
{
    /* tell the log stream to close its file */
    return log_str_->close_log_file();
}

#if 0 //$$$
/*
 *   This code is currently unused.  However, I'm leaving it in for now -
 *   the algorithm takes a little thought, so it would be nicer to be able
 *   to uncomment the existing code should we ever need it in the future.  
 */

/* ------------------------------------------------------------------------ */
/*
 *   Write UTF-8 text explicitly to the log file.  This can be used to add
 *   special text (such as prompt text) that would normally be suppressed
 *   from the log file.  When more mode is turned off, we don't
 *   automatically copy text to the log file; any text that the caller
 *   knows should be in the log file during times when more mode is turned
 *   off can be explicitly added with this function.
 *   
 *   If nl is true, we'll add a newline at the end of this text.  The
 *   caller should not include any newlines in the text being displayed
 *   here.  
 */
void CVmConsole::write_to_logfile(VMG_ const char *txt, int nl)
{
    /* if there's no log file, there's nothing to do */
    if (logfp_ == 0)
        return;

    /* 
     *   convert the text from UTF-8 to the local character set and write
     *   the converted text to the log file 
     */
    while (*txt != '\0')
    {
        char local_buf[128];
        size_t src_bytes_used;
        size_t out_bytes;

        /* convert as much as we can (leaving room for a null terminator) */
        out_bytes = G_cmap_to_ui->map_utf8(local_buf, sizeof(local_buf) - 1,
                                           txt, strlen(txt), &src_bytes_used);

        /* null-terminate the result */
        local_buf[out_bytes] = '\0';
        
        /* write the converted text */
        os_fprintf(logfp_, "%s", local_buf);

        /* skip the text we were able to convert */
        txt += src_bytes_used;
    }

    /* add a newline if desired */
    if (nl)
    {
        /* add a normal newline */
        os_fprintf(logfp_, "\n");

        /* if the logfile is an html target, write an HTML line break */
        if (log_str_->is_html_target() && log_str_->is_html_mode())
            os_fprintf(logfp_, "<BR HEIGHT=0>\n");
    }
}
#endif /* 0 */


/* ------------------------------------------------------------------------ */
/*
 *   Reset the MORE line counter.  This should be called whenever user
 *   input is read, since stopping to read user input makes it unnecessary
 *   to show another MORE prompt until the point at which input was
 *   solicited scrolls off the screen.  
 */
void CVmConsole::reset_line_count()
{
    /* reset the MORE counter in the display stream */
    disp_str_->reset_line_count();
}

/* ------------------------------------------------------------------------ */
/*
 *   Determine if HTML mode is active.  Returns true if so, false if not.
 *   Note that this merely indicates whether an "\H+" sequence is
 *   currently active -- this will return true after an "\H+" sequence,
 *   even on text-only interpreters.  
 */
int CVmConsole::is_html_mode()
{
    /* return the current HTML mode flag for the standard display stream */
    return disp_str_->is_html_mode();
}


/* ------------------------------------------------------------------------ */
/*
 *   Flush the output line.  We'll write to both the standard display and
 *   the log file, as needed.  
 */
void CVmConsole::flush(VMG_ vm_nl_type nl)
{
    /* flush the display stream */
    disp_str_->flush(vmg_ nl);

    /* flush the log stream, if we have an open log file */
    log_str_->flush(vmg_ nl);
}

/* ------------------------------------------------------------------------ */
/*
 *   Open a script file 
 */
void CVmConsole::open_script_file(const char *fname, int quiet,
                                  int script_more_mode)
{
    /* close any existing script file */
    close_script_file();

    /* open the new file */
    script_fp_ = osfoprt(fname, OSFTCMD);
    
    /* 
     *   if we successfully opened the file, remember the quiet setting;
     *   otherwise, we're definitely not reading a quiet script because
     *   we're not reading a script at all 
     */
    quiet_script_ = (script_fp_ != 0 && quiet);

    /*
     *   If we successfully opened a script file, remember the original
     *   MORE mode, and set the MORE mode that the caller wants in effect
     *   while processing the script.  If we didn't successfully open a
     *   script, don't make any change to the MORE mode.  
     */
    if (script_fp_ != 0)
        pre_script_more_mode_ = set_more_state(script_more_mode);
}

/*
 *   Close the script file 
 */
int CVmConsole::close_script_file()
{
    /* if we have a file, close it */
    if (script_fp_ != 0)
    {
        /* close the file */
        osfcls(script_fp_);

        /* forget the script file */
        script_fp_ = 0;

        /* 
         *   we're not reading any script any more, so forget any
         *   quiet-script mode that's in effect 
         */
        quiet_script_ = FALSE;

        /* 
         *   return the MORE mode in effect before we started reading the
         *   script file 
         */
        return pre_script_more_mode_;
    }
    else
    {
        /* 
         *   there's no script file - just return the current MORE mode,
         *   since we're not making any changes 
         */
        return is_more_mode();
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Open a command log file 
 */
int CVmConsole::open_command_log(const char *fname)
{
    /* close any existing command log file */
    close_command_log();
    
    /* remember the filename */
    strcpy(command_fname_, fname);

    /* open the file */
    command_fp_ = osfopwt(fname, OSFTCMD);

    /* return success if we successfully opened the file */
    return (command_fp_ == 0);
}

/* 
 *   close the active command log file 
 */
int CVmConsole::close_command_log()
{
    /* if there's a command log file, close it */
    if (command_fp_ != 0)
    {
        /* close the file */
        osfcls(command_fp_);

        /* set its file type */
        os_settype(command_fname_, OSFTCMD);

        /* forget the file */
        command_fp_ = 0;
    }

    /* success */
    return 0;
}


/* ------------------------------------------------------------------------ */
/*
 *   Read a line of input from the console.  Fills in the buffer with a
 *   null-terminated string in the UTF-8 character set.  Returns zero on
 *   success, non-zero on end-of-file.  
 */
int CVmConsole::read_line(VMG_ char *buf, size_t buflen)
{
    int old_more_mode;
    int retval;
    int echo_text;
    char local_buf[256];
    char *outp;
    size_t outlen;

    /* 
     *   presume we won't echo the text to the display; in most cases, it
     *   will be echoed to the display in the course of reading it from
     *   the keyboard 
     */
    echo_text = FALSE;
    
    /* 
     *   turn off MORE mode if it's on - we don't want a MORE prompt
     *   showing up in the midst of user input 
     */
    old_more_mode = set_more_state(FALSE);

    /* 
     *   flush the output; don't start a new line, since we might have
     *   displayed a prompt that is to be on the same line with the user
     *   input 
     */
    flush(vmg_ VM_NL_NONE);

    /* 
     *   reset the MORE line counter, since we're reading user input at
     *   the current point and shouldn't pause for a MORE prompt until the
     *   text we're reading has scrolled off the screen 
     */
    reset_line_count();

    /* if there's a script file, read from it */
    if (script_fp_ != 0)
    {
        /* try reading a line from the script file */
        if (read_line_from_script(local_buf, sizeof(local_buf)))
        {
            /* 
             *   End of script file - return to reading from the keyboard.
             *   The return value from close_script_file() is the MORE
             *   mode that was in effect before we started reading the
             *   script file; use this when we restore the enclosing MORE
             *   mode so that we restore the pre-script MORE mode when
             *   we return.
             */
            old_more_mode = close_script_file();

            /* turn off MORE mode while we're reading the command */
            set_more_state(FALSE);

            /* flush any output we generated while reading the script file */
            flush(vmg_ VM_NL_NONE);

            /* 
             *   reset the MORE line counter, since we might have
             *   generated output while reading the script file 
             */
            reset_line_count();
        }
        else
        {
            /* 
             *   we got a line from the script file - if we're not in
             *   quiet mode, make a note to echo the text to the display 
             */
            if (!quiet_script_)
                echo_text = TRUE;
        }

        /* success so far */
        retval = 0;
    }

    /* if we don't have a script file, read from the keyboard */
    if (script_fp_ == 0)
    {
        /* read a line from the keyboard */
        retval = (os_gets((uchar *)local_buf, sizeof(local_buf)) == 0);
    }

    /* restore the original MORE mode */
    set_more_state(old_more_mode);

    /* if we got an error, return it */
    if (retval != 0)
        return retval;

    /* 
     *   We got a command, or at least a partial command (we might have
     *   timed out, in which case we might still have a partial line in
     *   the buffer).  Write the input line, followed by a newline, to the
     *   log file.  Note that the text is still in the local character
     *   set, so we can write it directly to the log file.  
     */
    log_str_->disp_text_local_cs(local_buf);
    log_str_->disp_text_local_cs("\n");

    /*
     *   If we have a command log file, log the command (preceded by a ">"
     *   character, to indicate that it's a command, and followed by a
     *   newline) to the command log.  
     */
    if (command_fp_ != 0)
        os_fprintf(command_fp_, ">%s\n", local_buf);

    /* 
     *   Convert the text from the local UI character set to UTF-8.
     *   Reserve space in the output buffer for the null terminator. 
     */
    outp = buf;
    outlen = buflen - 1;
    G_cmap_from_ui->map(&outp, &outlen, local_buf, strlen(local_buf));

    /* add the null terminator */
    *outp = '\0';

    /* 
     *   If we need to echo the text (because we read it from a script
     *   file), do so now.
     */
    if (echo_text)
    {
        /* show the text */
        format_text(vmg_ buf);

        /* add a newline */
        format_text(vmg_ "\n");
    }

    /* return the result */
    return retval;
}

/*
 *   Read a line of text from the script file, if there is one.  Returns
 *   zero on success, non-zero if we reach the end of the script file or
 *   encounter any other error. 
 */
int CVmConsole::read_line_from_script(char *buf, size_t buflen)
{
    /* if there's no script file, return failure */
    if (script_fp_ == 0)
        return 1;

    /* keep going until we find a line that we like */
    for (;;)
    {
        char c;

        /* 
         *   Read the next character of input.  If it's not a newline,
         *   there's more text on the same line, so read the rest and
         *   determine what to do. 
         */
        c = osfgetc(script_fp_);
        if (c != '\n' && c != '\r')
        {
            /* there's more on this line - read the rest of the line */
            if (osfgets(buf, buflen, script_fp_) == 0)
            {
                /* end of file - return failure */
                return 1;
            }

            /* 
             *   if the line starts with '>', it's a command line;
             *   otherwise, it's a comment or something else, in which
             *   case we'll ignore it and keep looking for a line starting
             *   with '>' 
             */
            if (c == '>')
            {
                size_t len;
                
                /* 
                 *   if there are any trailing newline characters in the
                 *   buffer, remove them
                 */
                len = strlen(buf);
                while (len > 0
                       && (buf[len - 1] == '\n' || buf[len - 1] == '\r'))
                    buf[--len] = '\0';

                /* return success */
                return 0;
            }
        }
        else if (c == EOF)
        {
            /* end of file - return failure */
            return 1;
        }
    }
}

