#ifdef RCSID
static char RCSid[] =
"$Header$";
#endif

/* Copyright (c) 2000 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmgram.cpp - T3 Grammar Production metaclass
Function
  
Notes
  
Modified
  02/15/00 MJRoberts  - Creation
*/

#include <stdlib.h>
#include <string.h>

#include "t3std.h"
#include "vmtype.h"
#include "vmstack.h"
#include "vmlst.h"
#include "vmgram.h"
#include "vmerr.h"
#include "vmerrnum.h"
#include "vmmeta.h"
#include "vmdict.h"
#include "vmtobj.h"


/* ------------------------------------------------------------------------ */
/*
 *   Grammar production object - memory pool.  This is a simple
 *   suballocator that obtains large blocks from the system allocator,
 *   then hands out pieces of those blocks.  Allocating is very cheap
 *   (just a pointer increment in most cases), and we don't track
 *   individual allocations and deletions but just throw away all
 *   suballocated memory at once.  
 */

/* size of each memory block */
const size_t VMGRAMPROD_MEM_BLOCK_SIZE = 16*1024;

/*
 *   memory pool block 
 */
struct CVmGramProdMemBlock
{
    /* next block in chain */
    CVmGramProdMemBlock *nxt_;

    /* bytes of the block */
    char buf_[VMGRAMPROD_MEM_BLOCK_SIZE];
};

/*
 *   memory pool object 
 */
class CVmGramProdMem
{
public:
    CVmGramProdMem()
    {
        /* no memory blocks yet */
        block_head_ = 0;
        block_cur_ = 0;

        /* we don't yet have a block */
        cur_rem_ = 0;
        cur_free_ = 0;
    }

    ~CVmGramProdMem()
    {
        /* delete each of our blocks */
        while (block_head_ != 0)
        {
            CVmGramProdMemBlock *nxt;
            
            /* remember the next block */
            nxt = block_head_->nxt_;
            
            /* delete this block */
            t3free(block_head_);

            /* move on to the next */
            block_head_ = nxt;
        }
    }

    /* reset - delete all previously sub-allocated memory */
    void reset()
    {
        /* start over with the first block */
        block_cur_ = block_head_;

        /* initialize the free pointer for the new block, if we have one */
        if (block_cur_ != 0)
        {
            /* start at the beginning of the new block */
            cur_free_ = block_cur_->buf_;

            /* this entire block is available */
            cur_rem_ = VMGRAMPROD_MEM_BLOCK_SIZE;
        }
        else
        {
            /* there are no blocks, so there's no memory available yet */
            cur_rem_ = 0;
            cur_free_ = 0;
        }
    }

    /* allocate memory */
    void *alloc(size_t siz)
    {
        void *ret;
        
        /* round the size to the local hardware boundary */
        siz = osrndsz(siz);

        /* if it exceeds the block size, fail */
        if (siz > VMGRAMPROD_MEM_BLOCK_SIZE)
            return 0;

        /* if we don't have enough space in this block, go to the next */
        if (siz > cur_rem_)
        {
            /* allocate a new block if necessary */
            if (block_cur_ == 0)
            {
                /* there's nothing in the list yet - set up the first block */
                block_head_ = (CVmGramProdMemBlock *)
                              t3malloc(sizeof(CVmGramProdMemBlock));

                /* activate the new block */
                block_cur_ = block_head_;

                /* there's nothing after this block */
                block_cur_->nxt_ = 0;
            }
            else if (block_cur_->nxt_ == 0)
            {
                /* we're at the end of the list - add a block */
                block_cur_->nxt_ = (CVmGramProdMemBlock *)
                                   t3malloc(sizeof(CVmGramProdMemBlock));

                /* advance to the new block */
                block_cur_ = block_cur_->nxt_;

                /* the new block is the last in the chain */
                block_cur_->nxt_ = 0;
            }
            else
            {
                /* another block follows - advance to it */
                block_cur_ = block_cur_->nxt_;
            }

            /* start at the beginning of the new block */
            cur_free_ = block_cur_->buf_;

            /* this entire block is available */
            cur_rem_ = VMGRAMPROD_MEM_BLOCK_SIZE;
        }

        /* the return value will be the current free pointer */
        ret = cur_free_;

        /* consume the space */
        cur_free_ += siz;
        cur_rem_ -= siz;

        /* return the block location */
        return ret;
    }

private:
    /* head of block list */
    CVmGramProdMemBlock *block_head_;

    /* block we're currently suballocating out of */
    CVmGramProdMemBlock *block_cur_;

    /* current free pointer in current block */
    char *cur_free_;

    /* amount of space remaining in current block */
    size_t cur_rem_;
};

/* ------------------------------------------------------------------------ */
/*
 *   Token array entry.  We make a private copy of the token list during
 *   parsing using one of these structures for each token.  
 */
struct vmgramprod_tok
{
    /* token type */
    ulong typ_;

    /* token text */
    char *txt_;

    /* length of the token text */
    size_t len_;

    /* number of vocabulary properties associated with the word */
    size_t prop_cnt_;

    /* pointer to list of vocabulary properties associated with the word */
    vm_prop_id_t *props_;
};


/* ------------------------------------------------------------------------ */
/*
 *   Parsing state match object.  A match can be terminal, in which case
 *   it refers to a single token of input, or it can be a production, in
 *   which case it refers to a subtree of match objects.  
 */
struct CVmGramProdMatch
{
#if 0//$$$
    /* clone myself */
    CVmGramProdMatch *clone(CVmGramProdMem *mem)
    {
        CVmGramProdMatch *new_match;
        
        /* allocate a new item */
        new_match = (CVmGramProdMatch *)mem->alloc(sizeof(*new_match));

        /* copy my data */
        new_match->tok_pos_ = tok_pos_;
        new_match->prod_ = prod_;

        /* clone my subtree */
        new_state->prod_head_ = (prod_head_ != 0
                                 ? prod_head_->clone(mem) : 0);

        /* clone the rest of my list */
        new_state->nxt_ = (nxt_ != 0 ? nxt_->clone(mem) : 0);

        /* return the new item */
        return new_match;
    }
#endif

    /* allocate */
    static CVmGramProdMatch *alloc(CVmGramProdMem *mem, size_t tok_pos,
                                   vm_prop_id_t target_prop,
                                   vm_obj_id_t proc_obj,
                                   const CVmGramProdMatch **sub_match_list,
                                   size_t sub_match_cnt)
    {
        CVmGramProdMatch *match;
        
        /* allocate space */
        match = (CVmGramProdMatch *)mem->alloc(sizeof(CVmGramProdMatch));

        /* initialize */
        match->tok_pos_ = tok_pos;
        match->target_prop_ = target_prop;
        match->proc_obj_ = proc_obj;
        match->sub_match_list_ = sub_match_list;
        match->sub_match_cnt_ = sub_match_cnt;

        /* return the new match */
        return match;
    }

    /* property ID with which this match is associated */
    vm_prop_id_t target_prop_;
    
    /* 
     *   token position of the matching token; ignored if the production
     *   subtree information is valid 
     */
    size_t tok_pos_;

    /* 
     *   object ID of match processor - if this is valid, this match
     *   refers to a production; otherwise, the match refers to a single
     *   token 
     */
    vm_obj_id_t proc_obj_;

    /* pointer to array of sub-matches */
    const CVmGramProdMatch **sub_match_list_;

    /* number of sub-match entries */
    size_t sub_match_cnt_;
};

/*
 *   Top-level match list holder 
 */
struct CVmGramProdMatchEntry
{
    /* the match tree */
    CVmGramProdMatch *match_;

    /* token position at the end of the match */
    size_t tok_pos_;

    /* next entry in the list */
    CVmGramProdMatchEntry *nxt_;
};

/*
 *   Parsing state object.  At any given time, we will have one or more of
 *   these state objects in our processing queue.  A state object tracks a
 *   parsing position - the token position, the alternative position, the
 *   match list so far for the alternative (i.e., the match for each item
 *   in the alternative up to but not including the current position), and
 *   the enclosing parsing state object (i.e., the parsing state that
 *   "recursed" into this alternative).  
 */
struct CVmGramProdState
{
    /* allocate */
    static CVmGramProdState *alloc(CVmGramProdMem *mem,
                                   int tok_pos, int alt_pos,
                                   CVmGramProdState *enclosing,
                                   const char *altp, vm_obj_id_t prod_obj,
                                   int circular_alt)
    {
        size_t match_cnt;
        CVmGramProdState *state;

        /* get the number of match list slots we'll need */
        match_cnt = vmgram_alt_tokcnt(altp);

        /* allocate space, including space for the match list */
        state = (CVmGramProdState *)
                mem->alloc(sizeof(CVmGramProdState)
                           + (match_cnt - 1)*sizeof(CVmGramProdMatch));

        /* initialize */
        state->init(tok_pos, alt_pos, enclosing, altp, prod_obj,
                    circular_alt);

        /* return the new item */
        return state;
    }
    
    /* initialize */
    void init(int tok_pos, int alt_pos, CVmGramProdState *enclosing,
              const char *altp, vm_obj_id_t prod_obj, int circular_alt)
    {
        /* remember the position parameters */
        tok_pos_ = tok_pos;
        alt_pos_ = alt_pos;

        /* remember our stack position */
        enclosing_ = enclosing;

        /* remember our alternative image data pointer */
        altp_ = altp;

        /* remember the production object */
        prod_obj_ = prod_obj;

        /* note if we had circular alternatives in the same production */
        circular_alt_ = circular_alt;

        /* no target property for sub-production yet */
        sub_target_prop_ = VM_INVALID_PROP;

        /* we're not in the queue yet */
        nxt_ = 0;
    }

    /* clone the state */
    CVmGramProdState *clone(CVmGramProdMem *mem)
    {
        CVmGramProdState *new_state;
        CVmGramProdState *new_enclosing;

        /* clone our enclosing state */
        new_enclosing = (enclosing_ != 0 ? enclosing_->clone(mem) : 0);
        
        /* create a new state object */
        new_state = alloc(mem, tok_pos_, alt_pos_, new_enclosing,
                          altp_, prod_obj_, circular_alt_);

        /* copy the target sub-production property */
        new_state->sub_target_prop_ = sub_target_prop_;

        /* copy the valid portion of my match list */
        if (alt_pos_ != 0)
            memcpy(new_state->match_list_, match_list_,
                   alt_pos_ * sizeof(match_list_[0]));

        /* we're not yet in any queue */
        new_state->nxt_ = 0;

        /* return the clone */
        return new_state;
    }

    /* current token position, as an index into the token list */
    size_t tok_pos_;

    /* 
     *   current alternative position, as an index into the alternative's
     *   list of items 
     */
    size_t alt_pos_;

    /* 
     *   the enclosing state object - this is the state object that
     *   "recursed" to our position 
     */
    CVmGramProdState *enclosing_;

    /* pointer to alternative definition in image file data */
    const char *altp_;

    /* object ID of the production object */
    vm_obj_id_t prod_obj_;

    /* the target property for the pending sub-production */
    vm_prop_id_t sub_target_prop_;

    /* next production state in the processing queue */
    CVmGramProdState *nxt_;

    /* the production contains circular alternatives */
    int circular_alt_;

    /* 
     *   Match list.  This list is allocated with enough space for one
     *   match for each of our items (the number of items is given by
     *   vmgram_alt_tokcnt(altp_)).  At any given time, these will be
     *   valid up to but not including the current alt_pos_ index.  
     */
    const struct CVmGramProdMatch *match_list_[1];
};

/*
 *   Queues.  We maintain three queues: the primary work queue, which
 *   contains pending parsing states that we're still attempting to match;
 *   the success queue, which contains a list of match trees for
 *   successfully completed matches; and the badness queue, which contains
 *   a list of parsing states that we can try as fallbacks if we fail to
 *   find a match in any of the work queue items.  
 */
struct CVmGramProdQueue
{
    /* clear the queues */
    void clear()
    {
        work_queue_ = 0;
        badness_queue_ = 0;
        success_list_ = 0;
    }

    /* head of work queue */
    CVmGramProdState *work_queue_;

    /* head of badness queue */
    CVmGramProdState *badness_queue_;

    /* head of success list */
    CVmGramProdMatchEntry *success_list_;
};

/* ------------------------------------------------------------------------ */
/*
 *   Grammar-Production object statics 
 */

/* metaclass registration object */
static CVmMetaclassGramProd metaclass_reg_obj;
CVmMetaclass *CVmObjGramProd::metaclass_reg_ = &metaclass_reg_obj;

/* function table */
int (CVmObjGramProd::
     *CVmObjGramProd::func_table_[])(VMG_ vm_obj_id_t self,
                                     vm_val_t *retval, uint *argc) =
{
    &CVmObjGramProd::getp_undef,
    &CVmObjGramProd::getp_parse
};

/* ------------------------------------------------------------------------ */
/*
 *   Grammar-Production metaclass implementation 
 */

/* 
 *   create dynamically using stack arguments 
 */
vm_obj_id_t CVmObjGramProd::create_from_stack(VMG_ const uchar **pc_ptr,
                                              uint argc)
{
    /* this type of object cannot be dynamically created */
    err_throw(VMERR_BAD_DYNAMIC_NEW);

    /* we won't reach this point, but the compiler wouldn't know */
    return VM_INVALID_OBJ;
}

/*
 *   constructor 
 */
CVmObjGramProd::CVmObjGramProd(VMG0_)
{
    /* allocate our extension structure from the variable heap */
    ext_ = (char *)G_mem->get_var_heap()
           ->alloc_mem(sizeof(vm_gram_ext), this);

    /* we have no image data yet */
    get_ext()->image_data_ = 0;
    get_ext()->image_data_size_ = 0;

    /* create our memory pool */
    get_ext()->mem_ = new CVmGramProdMem();

    /* 
     *   allocate initial property enumeration space (we'll expand this as
     *   needed, so the initial size is just a guess) 
     */
    get_ext()->prop_enum_max_ = 64;
    get_ext()->prop_enum_arr_ = (vm_prop_id_t *)
                                t3malloc(get_ext()->prop_enum_max_
                                         * sizeof(vm_prop_id_t));
}

/* 
 *   notify of deletion 
 */
void CVmObjGramProd::notify_delete(VMG_ int in_root_set)
{
    /* free our additional data */
    if (ext_ != 0)
    {
        /* delete our memory pool */
        delete get_ext()->mem_;

        /* delete our property enumeration space */
        t3free(get_ext()->prop_enum_arr_);
        
        /* free the extension */
        G_mem->get_var_heap()->free_mem(ext_);
    }
}

/* 
 *   get a property 
 */
int CVmObjGramProd::get_prop(VMG_ vm_prop_id_t prop, vm_val_t *retval,
                             vm_obj_id_t self, vm_obj_id_t *source_obj,
                             uint *argc)
{
    ushort func_idx;
    
    /* if we succeed, the result will come directly from this object */
    *source_obj = self;

    /* translate the property index to an index into our function table */
    func_idx = G_meta_table
               ->prop_to_vector_idx(metaclass_reg_->get_reg_idx(), prop);

    /* call the appropriate function */
    if ((this->*func_table_[func_idx])(vmg_ self, retval, argc))
        return TRUE;

    /* inherit default handling */
    return CVmObject::get_prop(vmg_ prop, retval, self, source_obj, argc);
}

/* 
 *   set a property 
 */
void CVmObjGramProd::set_prop(VMG_ class CVmUndo *undo,
                              vm_obj_id_t self, vm_prop_id_t prop,
                              const vm_val_t *val)
{
    /* no settable properties - throw an error */
    err_throw(VMERR_INVALID_SETPROP);
}

/* 
 *   load from an image file 
 */
void CVmObjGramProd::load_from_image(VMG_ vm_obj_id_t,
                                     const char *ptr, size_t siz)
{
    /* remember where our image data comes from */
    get_ext()->image_data_ = ptr;
    get_ext()->image_data_size_ = siz;
}

/*
 *   Context for enum_props_cb 
 */
struct enum_props_ctx
{
    /* object extension */
    vm_gram_ext *ext;
    
    /* number of properties in our list so far */
    size_t cnt;
};

/*
 *   Callback for enumerating word properties 
 */
void CVmObjGramProd::enum_props_cb(VMG_ void *ctx0, vm_prop_id_t prop)
{
    enum_props_ctx *ctx = (enum_props_ctx *)ctx0;
    size_t i;

    /* if this property is already in our list, don't add it again */
    for (i = 0 ; i < ctx->cnt ; ++i)
    {
        /* if this one matches, we can return immediately */
        if (ctx->ext->prop_enum_arr_[i] == prop)
            return;
    }
    
    /* we need to add it - if the array is full, expand it */
    if (ctx->cnt == ctx->ext->prop_enum_max_)
    {
        /* reallocate the array with more space */
        ctx->ext->prop_enum_max_ += 64;
        ctx->ext->prop_enum_arr_ = (vm_prop_id_t *)
                                   t3realloc(ctx->ext->prop_enum_arr_,
                                             ctx->ext->prop_enum_max_
                                             * sizeof(vm_prop_id_t));
    }

    /* add the property to the array */
    ctx->ext->prop_enum_arr_[ctx->cnt] = prop;

    /* count the new entry */
    ctx->cnt++;
}

/*
 *   Execute the parseToken method 
 */
int CVmObjGramProd::getp_parse(VMG_ vm_obj_id_t self,
                               vm_val_t *retval, uint *argc)
{
    vm_val_t *tokval;
    vm_val_t *typval;
    vm_val_t dictval;
    CVmObjDict *dict;
    size_t tok_cnt;
    size_t tok_cnt2;
    size_t i;
    vmgramprod_tok *tok;
    const char *p;
    int orig_argc = (argc != 0 ? *argc : 0);
    CVmGramProdQueue queues;
    CVmObjList *lst;
    size_t succ_cnt;
    CVmGramProdMatchEntry *match;
    size_t trunc_len;
    
    /* check arguments */
    if (get_prop_check_argc(retval, argc, 3))
        return TRUE;

    /* presume we won't allow any truncated matches */
    trunc_len = 0;

    /* reset our memory pool */
    get_ext()->mem_->reset();

    /* clear the work queues */
    queues.clear();

    /* 
     *   get the tokenList argument and make sure it's a list; leave it on
     *   the stack for now so it remains reachable for the garbage
     *   collector 
     */
    tokval = G_stk->get(0);
    if (tokval->get_as_list(vmg0_) == 0)
        err_throw(VMERR_LIST_VAL_REQD);

    /* get the tokenTypeList argument and make sure it's a list */
    typval = G_stk->get(1);
    if (typval->get_as_list(vmg0_) == 0)
        err_throw(VMERR_LIST_VAL_REQD);

    /* check for a dictionary argument */
    if (orig_argc >= 3)
    {
        /* get the dictionary argument */
        dictval = *G_stk->get(2);

        /* make sure it's an object or nil */
        if (dictval.typ != VM_OBJ && dictval.typ != VM_NIL)
            err_throw(VMERR_OBJ_VAL_REQD);
    }
    else
    {
        /* use a nil dictionary by default */
        dictval.set_nil();
    }

    /* get the dictionary object and check that it's really a dictionary */
    if (dictval.typ != VM_NIL)
    {
        /* if it's not a CVmObjDict object, it's the wrong type */
        if (!CVmObjDict::is_dictionary_obj(vmg_ dictval.val.obj))
        {
            /* wrong type - throw an error */
            err_throw(VMERR_INVAL_OBJ_TYPE);
        }

        /* get the VM object pointer */
        dict = (CVmObjDict *)vm_objp(vmg_ dictval.val.obj);

        /* note the dictionary's truncatino length */
        trunc_len = dict->get_trunc_len();
    }
    else
    {
        /* there's no dictionary object */
        dict = 0;
    }

    /* get the lengths of the two lists, and use the shorter of the two */
    tok_cnt = vmb_get_len(tokval->get_as_list(vmg0_));
    tok_cnt2 = vmb_get_len(typval->get_as_list(vmg0_));
    if (tok_cnt2 < tok_cnt)
        tok_cnt = tok_cnt2;

    /* 
     *   For quick and easy access, make our own private copy of the token
     *   and type lists.  First, allocate an array of token structures.  
     */
    tok = (vmgramprod_tok *)get_ext()->mem_
          ->alloc(tok_cnt * sizeof(vmgramprod_tok));

    /* get the type list pointer */
    p = typval->get_as_list(vmg0_);

    /* copy the token types into the list */
    for (i = 0 ; i < tok_cnt ; ++i)
    {
        vm_val_t val;

        /* get this element from the list */
        CVmObjList::index_list(vmg_ &val, p, i+1);

        /* use the value if it's an enumerator; if not, use zero */
        if (val.typ == VM_ENUM)
            tok[i].typ_ = val.val.enumval;
        else
            tok[i].typ_ = 0;

        /* presume we won't have any vocabulary properties for the word */
        tok[i].prop_cnt_ = 0;
        tok[i].props_ = 0;
    }

    /* copy the token strings */
    for (i = 0 ; i < tok_cnt ; ++i)
    {
        vm_val_t val;

        /* translate the list pointer again in case of swapping */
        p = tokval->get_as_list(vmg0_);

        /* get the value */
        CVmObjList::index_list(vmg_ &val, p, i+1);

        /* get the string value */
        p = val.get_as_string(vmg0_);

        /* if it's a string, copy it; otherwise, ignore the value */
        if (p != 0)
        {
            char *new_str;
            
            /* allocate space for a copy of the string */
            new_str = (char *)get_ext()->mem_->alloc(vmb_get_len(p));
            if (new_str == 0)
            {
                /* the string is too long - ignore it */
                tok[i].txt_ = 0;
                tok[i].len_ = 0;
            }
            else
            {
                /* copy the string */
                memcpy(new_str, p + VMB_LEN, vmb_get_len(p));

                /* point the token to our copy of the string */
                tok[i].txt_ = new_str;
                tok[i].len_ = vmb_get_len(p);

                /* 
                 *   if we have a dictionary, enumerate the properties
                 *   associated with the word 
                 */
                if (dict != 0)
                {
                    enum_props_ctx ctx;
                    
                    /* set up our callback context */
                    ctx.ext = get_ext();
                    ctx.cnt = 0;

                    /* enumerate the properties */
                    dict->enum_word_props(vmg_ &enum_props_cb, &ctx,
                                          tok[i].txt_, tok[i].len_, TRUE);

                    /* 
                     *   if we found any properties for the word, make a
                     *   copy for the token list 
                     */
                    if (ctx.cnt != 0)
                    {
                        /* allocate space for the list */
                        tok[i].prop_cnt_ = ctx.cnt;
                        tok[i].props_ =
                            (vm_prop_id_t *)get_ext()->mem_
                            ->alloc(ctx.cnt * sizeof(vm_prop_id_t));

                        /* copy the list */
                        memcpy(tok[i].props_, get_ext()->prop_enum_arr_,
                               ctx.cnt * sizeof(vm_prop_id_t));
                    }
                }
            }
        }
        else
        {
            /* it's not a string - use a null string value for this token */
            tok[i].txt_ = 0;
            tok[i].len_ = 0;
        }
    }

    /* enqueue a match state item for each of our alternatives */
    enqueue_alts(vmg_ get_ext()->mem_, tok, tok_cnt, 0, 0,
                 &queues, self, FALSE, 0, trunc_len);

    /* process the work queue */
    process_work_queue(vmg_ get_ext()->mem_, tok, tok_cnt,
                       &queues, trunc_len);

    /* count the entries in the success list */
    for (succ_cnt = 0, match = queues.success_list_ ; match != 0 ;
         ++succ_cnt, match = match->nxt_) ;

    /* allocate the return list */
    retval->set_obj(CVmObjList::create(vmg_ FALSE, succ_cnt));
    lst = (CVmObjList *)vm_objp(vmg_ retval->val.obj);

    /* initially clear the elements of the return list to nil */
    for (i = 0 ; i < succ_cnt ; ++i)
    {
        vm_val_t nil_val;

        /* set this element to nil */
        nil_val.set_nil();
        lst->cons_set_element(i, &nil_val);
    }

    /* push the list momentarily to protect it from garbage collection */
    G_stk->push(retval);

    /* set the elements */
    for (succ_cnt = 0, match = queues.success_list_ ; match != 0 ;
         ++succ_cnt, match = match->nxt_)
    {
        vm_val_t tree_val;
        vm_val_t sublst_val;
        CVmObjList *sublst;
        vm_val_t ele_val;
        
        /* build the object tree for this success object */
        build_match_tree(vmg_ match->match_, tokval, &tree_val);

        /* allocate the list to contain [tokensUsed, matchTree] */
        sublst_val.set_obj(CVmObjList::create(vmg_ FALSE, 2));
        sublst = (CVmObjList *)vm_objp(vmg_ sublst_val.val.obj);

        /* set the first element to the next token index */
        ele_val.set_int(match->tok_pos_);
        sublst->cons_set_element(0, &ele_val);

        /* set the second element to the match tree */
        sublst->cons_set_element(1, &tree_val);
        
        /* 
         *   add the sublist to the main return list (note that this
         *   protects it from garbage collection by virtue of the main
         *   return list being protected from garbage collection) 
         */
        lst->cons_set_element(succ_cnt, &sublst_val);
    }

    /* discard the gc protection */
    G_stk->discard();

    /* discard arguments */
    G_stk->discard(orig_argc);

    /* evaluation successful */
    return TRUE;
}

/*
 *   Build an object tree for a match 
 */
void CVmObjGramProd::build_match_tree(VMG_ const CVmGramProdMatch *match,
                                      vm_val_t *toklist, vm_val_t *retval)
{
    /* check to see what kind of match we have */
    if (match->proc_obj_ != VM_INVALID_OBJ)
    {
        vm_obj_id_t obj_id;
        CVmObjTads *objp;
        size_t i;
        vm_val_t val;

        /* create the object to hold the current tree level */
        obj_id = CVmObjTads::create(vmg_ FALSE, 1, VMTOBJ_PROP_INIT);
        objp = (CVmObjTads *)vm_objp(vmg_ obj_id);

        /* set the superclass to the match's processor object */
        objp->set_sc(0, match->proc_obj_);

        /* push this object to protect it from gc for a moment */
        val.set_obj(obj_id);
        G_stk->push(&val);

        /* build the sub-match entries */
        for (i = 0 ; i < match->sub_match_cnt_ ; ++i)
        {
            /* build this entry */
            build_match_tree(vmg_ match->sub_match_list_[i], toklist, &val);

            /* set the processor object property for the value */
            if (match->sub_match_list_[i]->target_prop_ != VM_INVALID_PROP)
                objp->set_prop(vmg_ G_undo, obj_id,
                               match->sub_match_list_[i]->target_prop_,
                               &val);
        }

        /* discard our gc protection */
        G_stk->discard();

        /* the return value is the object */
        retval->set_obj(obj_id);
    }
    else
    {
        const char *lstp;

        /* get the token list */
        lstp = toklist->get_as_list(vmg0_);

        /* make sure the index is in range */
        if (match->tok_pos_ < vmb_get_len(lstp))
        {
            /* get the token value from the list */
            CVmObjList::index_list(vmg_ retval, lstp, match->tok_pos_ + 1);
        }
        else
        {
            /* 
             *   the index is past the end of the list - this must be a
             *   '*' token that matched nothing (i.e., the token list was
             *   fully consumed before we reached the '*'), so just return
             *   nil for the match value 
             */
            retval->set_nil();
        }
    }
}

/*
 *   Process the work queue 
 */
void CVmObjGramProd::process_work_queue(VMG_ CVmGramProdMem *mem,
                                        const vmgramprod_tok *tok,
                                        size_t tok_cnt,
                                        CVmGramProdQueue *queues,
                                        size_t trunc_len)
{
    /* keep going until the work queue and badness queue are empty */
    for (;;)
    {
        /* 
         *   If the work queue is empty, fall back on the badness queue.
         *   Ignore the badness queue if we have any successful matches,
         *   since non-badness matches always take precedence over badness
         *   items. 
         */
        if (queues->work_queue_ == 0 && queues->success_list_ == 0)
        {
            int first;
            int min_badness;
            CVmGramProdState *cur;
            CVmGramProdState *prv;
            CVmGramProdState *nxt;
            
            /* find the lowest badness rating in the queue */
            for (first = TRUE, cur = queues->badness_queue_ ; cur != 0 ;
                 cur = cur->nxt_, first = FALSE)
            {
                /* 
                 *   if we're on the first item, note its badness as the
                 *   tentative minimum; otherwise, note the current item's
                 *   badness if it's lower than the lowest we've seen so
                 *   far 
                 */
                if (first || vmgram_alt_badness(cur->altp_) < min_badness)
                    min_badness = vmgram_alt_badness(cur->altp_);
            }

            /* 
             *   move all of the items whose badness matches the minimum
             *   badness out of the badness queue and into the work queue 
             */
            for (cur = queues->badness_queue_, prv = 0 ; cur != 0 ;
                 cur = nxt)
            {
                /* remember the next item, in case we remove this one */
                nxt = cur->nxt_;
                
                /* if this item has minimum badness, move it */
                if (vmgram_alt_badness(cur->altp_) == min_badness)
                {
                    /* unlink it from the badness queue */
                    if (prv == 0)
                        queues->badness_queue_ = cur->nxt_;
                    else
                        prv->nxt_ = cur->nxt_;

                    /* link it into the work queue */
                    cur->nxt_ = queues->work_queue_;
                    queues->work_queue_ = cur;

                    /* 
                     *   stop here - we just want to process one badness
                     *   item at time
                     */
                    break;
                }
                else
                {
                    /* 
                     *   this item is staying in the list - note it as the
                     *   item preceding the next item 
                     */
                    prv = cur;
                }
            }
        }

        /* if the work queue is still empty, we're out of work to do */
        if (queues->work_queue_ == 0)
            break;
        
        /* process the head of the work queue */
        process_work_queue_head(vmg_ mem, tok, tok_cnt, queues, trunc_len);
    }
}

/*
 *   Process a work queue entry 
 */
void CVmObjGramProd::process_work_queue_head(VMG_ CVmGramProdMem *mem,
                                             const vmgramprod_tok *tok,
                                             size_t tok_cnt,
                                             CVmGramProdQueue *queues,
                                             size_t trunc_len)
{
    CVmGramProdState *state;
    CVmGramProdMatch *match;
    const char *tokp;
    size_t i;
    vm_prop_id_t enclosing_target_prop;
    int matched_star;
    
    /* get the first entry from the queue */
    state = queues->work_queue_;

    /* unlink this entry from the queue */
    queues->work_queue_ = state->nxt_;
    state->nxt_ = 0;

    /* get the token pointer for the next active entry in the alternative */
    for (tokp = vmgram_alt_tokptr(state->altp_), i = 0 ;
         i < state->alt_pos_ ;
         tokp = get_next_alt_tok(tokp), ++i) ;

    /* presume we won't match a '*' token */
    matched_star = FALSE;

    /* process the remaining items in the entry */
    while (state->alt_pos_ < vmgram_alt_tokcnt(state->altp_))
    {
        int match;

        /* see what we have for this token */
        switch(vmgram_tok_type(tokp))
        {
        case VMGRAM_MATCH_PROD:
            {
                vm_obj_id_t sub_obj_id;
                CVmObjGramProd *sub_objp;

                /*
                 *   This is a sub-production node.  Get the
                 *   sub-production object.  
                 */
                sub_obj_id = vmgram_tok_prod_obj(tokp);

                /* make sure it's of the correct metaclass */
                if (!CVmObjGramProd::is_gramprod_obj(vmg_ sub_obj_id))
                {
                    /* wrong type - throw an error */
                    err_throw(VMERR_INVAL_OBJ_TYPE);
                }

                /* get the sub-production object */
                sub_objp = (CVmObjGramProd *)vm_objp(vmg_ sub_obj_id);

                /* 
                 *   set my subproduction target property, so that the
                 *   sub-production can set the target property correctly 
                 */
                state->sub_target_prop_ = vmgram_tok_prop(tokp);
                
                /* enqueue the alternatives for the sub-production */
                sub_objp->enqueue_alts(vmg_ mem, tok, tok_cnt,
                                       state->tok_pos_, state, queues,
                                       sub_obj_id, FALSE, 0, trunc_len);
            }

            /* 
             *   Do not process the current state any further for now -
             *   we'll get back to it when (and if) we finish processing
             *   the sub-production.  Note that we don't even put the
             *   current state back in the queue - it's stacked behind the
             *   sub-production, and will be re-enqueued when we
             *   successfully finish with the sub-production.  
             */
            return;
            
        case VMGRAM_MATCH_SPEECH:
            /* part of speech - check to see if the current token matches */
            match = (state->tok_pos_ < tok_cnt
                     && find_prop_in_tok(&tok[state->tok_pos_],
                                         vmgram_tok_voc_prop(tokp)));
            break;
            
        case VMGRAM_MATCH_LITERAL:
            /* literal - check for a match to the string */
            match = (state->tok_pos_ < tok_cnt
                     && tok_equals_lit(tok[state->tok_pos_].txt_,
                                       tok[state->tok_pos_].len_,
                                       vmgram_tok_lit_txt(tokp),
                                       vmgram_tok_lit_len(tokp),
                                       trunc_len));
            break;
            
        case VMGRAM_MATCH_TOKTYPE:
            /* token type */
            match = (state->tok_pos_ < tok_cnt
                     && (tok[state->tok_pos_].typ_
                         == vmgram_tok_tok_enum(tokp)));
            break;

        case VMGRAM_MATCH_STAR:
            /* 
             *   this matches anything remaining (it also matches if
             *   there's nothing remaining) 
             */
            match = TRUE;

            /* note that we matched a star */
            matched_star = TRUE;
            break;
        }

        /* check for a match */
        if (match)
        {
            /* 
             *   we matched this token - add a token match to our state's
             *   match list for the current position 
             */
            state->match_list_[state->alt_pos_] =
                CVmGramProdMatch::alloc(mem, state->tok_pos_,
                                        vmgram_tok_prop(tokp),
                                        VM_INVALID_OBJ, 0, 0);

            /* we're done with this alternative item - move on */
            state->alt_pos_++;

            /* consume the input token if we matched a real token */
            if (!matched_star)
                state->tok_pos_++;
            
            /* move on to the next alternative token item */
            tokp = get_next_alt_tok(tokp);
        }
        else
        {
            /* 
             *   This is not a match - reject this alternative.  We do not
             *   need to process this item further, so we can stop now,
             *   without returning this state to the work queue.  
             */

            /* abandon this work queue item */
            return;
        }
    }

    /* 
     *   If we make it here, we've reached the end of this alternative and
     *   have matched everything in it.  This means that the alternative
     *   was successful, so we can perform a 'reduce' operation to replace
     *   the matched tokens with this production.  
     */

    /* if we have an enclosing state, get its target property */
    enclosing_target_prop = (state->enclosing_ != 0
                             ? state->enclosing_->sub_target_prop_
                             : VM_INVALID_PROP);

    /* create a match for the alternative */
    match = CVmGramProdMatch::alloc(mem, 0, enclosing_target_prop,
                                    vmgram_alt_procobj(state->altp_),
                                    &state->match_list_[0],
                                    vmgram_alt_tokcnt(state->altp_));

    /*
     *   If the state we just matched was marked on creation as coming
     *   from an alternative with circular alternatives, we've just come
     *   up with a match for each circular alternative.  To avoid infinite
     *   recursion, we never enqueue the circular alternatives; however,
     *   now that we know we have a match for the first element of each
     *   circular alternative, we can check each of them to see if we can
     *   enqueue them.
     */
    if (state->circular_alt_)
    {
        vm_obj_id_t prod_obj_id;
        CVmObjGramProd *prod_objp;

        /* 
         *   get the production from which this alternative came (there
         *   should be no need to check the metaclass, since we could only
         *   have created the state object from a valid production in the
         *   first place) 
         */
        prod_obj_id = state->prod_obj_;
        prod_objp = (CVmObjGramProd *)vm_objp(vmg_ prod_obj_id);

        /* 
         *   Enqueue the matching circular alternatives from the original
         *   production.  Our match becomes the first token match in the
         *   circular alternative (i.e., it matches the leading circular
         *   reference element).  
         */
        prod_objp->enqueue_alts(vmg_ mem, tok, tok_cnt, state->tok_pos_,
                                state->enclosing_, queues, prod_obj_id,
                                TRUE, match, trunc_len);
    }

    /* check for an enclosing state to pop */
    if (state->enclosing_ != 0)
    {
        /* add the match to the enclosing state's match list */
        state->enclosing_->match_list_[state->enclosing_->alt_pos_] = match;
        state->enclosing_->alt_pos_++;

        /* 
         *   move the enclosing state's token position to our token
         *   position - since it now encompasses our match, it has
         *   consumed all of the tokens therein 
         */
        state->enclosing_->tok_pos_ = state->tok_pos_;

        /* enqueue the enclosing state so we can continue parsing it */
        enqueue_state(state->enclosing_, queues);
    }
    else
    {
        CVmGramProdMatchEntry *entry;
        
        /* 
         *   This is a top-level state, so we've completed parsing.  If
         *   we've consumed all input, or we matched a '*' token, we were
         *   successful.  If there's more input remaining, this is not a
         *   match.  
         */
        if (state->tok_pos_ < tok_cnt && !matched_star)
        {
            /*
             *   There's more input remaining, so this isn't a match.
             *   Reject this alternative.  We do not need to process this
             *   item further, so stop now without returning this state
             *   object to the work queue. 
             */

            /* abandon this work queue item */
            return;
        }

        /* create a new entry for the match list */
        entry = (CVmGramProdMatchEntry *)mem->alloc(sizeof(*entry));
        entry->match_ = match;
        entry->tok_pos_ = state->tok_pos_;

        /* link the entry into the list */
        entry->nxt_ = queues->success_list_;
        queues->success_list_ = entry;
    }
}

/*
 *   Visit my alternatives and enqueue each one that is a possible match
 *   for a given token list.  
 */
void CVmObjGramProd::enqueue_alts(VMG_ CVmGramProdMem *mem,
                                  const vmgramprod_tok *tok,
                                  size_t tok_cnt, size_t start_tok_pos,
                                  CVmGramProdState *enclosing_state,
                                  CVmGramProdQueue *queues, vm_obj_id_t self,
                                  int circ_only,
                                  CVmGramProdMatch *circ_match,
                                  size_t trunc_len)
{
    size_t alt_cnt;
    size_t i;
    const char *image_ptr;
    const char *altp;
    const char *p;
    int need_to_clone;
    int found_circ;

    /* 
     *   We don't need to clone the enclosing state until we've set up one
     *   new state referring back to it.  However, if we're doing circular
     *   alternatives, the enclosing state could also be enqueued
     *   separately as its own state, so do clone all required copies in
     *   this case.  
     */
    need_to_clone = circ_only;

    /* get our image file data pointer */
    image_ptr = get_ext()->image_data_;

    /* get the number of alternatives and set up at the first one */
    alt_cnt = osrp2(image_ptr);
    altp = image_ptr + 2;

    /* scan the alternatives for direct circular references */
    for (found_circ = FALSE, i = 0, p = altp ; i < alt_cnt ; ++i)
    {
        const char *tokp;
        size_t alt_tok_cnt;

        /* get the token count and first token for this alternative */
        alt_tok_cnt = vmgram_alt_tokcnt(p);
        tokp = vmgram_alt_tokptr(p);

        /* if the alternative is circular, note it */
        if (alt_tok_cnt != 0
            && vmgram_tok_type(tokp) == VMGRAM_MATCH_PROD
            && vmgram_tok_prod_obj(tokp) == self)
        {
            /* note the existence of a circular reference */
            found_circ = TRUE;

            /* 
             *   we only need to know that there's at least one, so
             *   there's no need to look any further 
             */
            break;
        }

        /* skip the tokens */
        for ( ; alt_tok_cnt != 0 ;
              --alt_tok_cnt, tokp = get_next_alt_tok(tokp)) ;

        /* the next alternative starts after the last token */
        p = tokp;
    }

    /* 
     *   run through our alternatives and enqueue each one that looks
     *   plausible 
     */
    for (i = 0 ; i < alt_cnt ; ++i)
    {
        const char *first_tokp;
        const char *tokp;
        size_t alt_tok_cnt;
        size_t cur_alt_tok;
        size_t tok_idx;
        int found_mismatch;
        int prod_before;
        int is_circ;
        
        /* start at the first token remaining in the string */
        tok_idx = start_tok_pos;

        /* get the token count for this alternative */
        alt_tok_cnt = vmgram_alt_tokcnt(altp);

        /* get the first token for this alternative */
        tokp = first_tokp = vmgram_alt_tokptr(altp);
        
        /* presume we will find no mismatches in this check */
        found_mismatch = FALSE;

        /* 
         *   note whether this alternative is a direct circular reference
         *   or not (it is directly circular if the first token points
         *   back to this same production) 
         */
        is_circ = (alt_tok_cnt != 0
                   && vmgram_tok_type(tokp) == VMGRAM_MATCH_PROD
                   && vmgram_tok_prod_obj(tokp) == self);
        
        /* 
         *   we don't have a production before the current item (this is
         *   important because it determines whether we must consider the
         *   current token or any following token when trying to match a
         *   literal, part of speech, or token type) 
         */
        prod_before = FALSE;

        /* scan the tokens in the alternative */
        for (cur_alt_tok = 0 ; cur_alt_tok < alt_tok_cnt ;
             ++cur_alt_tok, tokp = get_next_alt_tok(tokp))
        {
            int match;
            int match_star;
            int ignore_item;

            /* presume we won't find a match */
            match = FALSE;
            match_star = FALSE;

            /* presume we won't be able to ignore the item */
            ignore_item = FALSE;

            /* 
             *   if we're at the last token, and this is a 'star' item,
             *   match all remaining tokens 
             */
            if (tok_idx == tok_cnt
                && vmgram_tok_type(tokp) == VMGRAM_MATCH_STAR)
            {
                /* 
                 *   it's a match, because '*' matches anything - even
                 *   nothing, which is what we have left 
                 */
                match = TRUE;
                match_star = TRUE;
            }

            /* 
             *   Try to find a match to the current alternative item.  If
             *   we've already found a mismatch, don't boether, since we
             *   need no further evidence to skip this alternative; in
             *   this case we're still looping over the alternative items
             *   simply to find the beginning of the next alternative. 
             */
            for ( ; tok_idx < tok_cnt && !found_mismatch ; ++tok_idx)
            {
                /* see what kind of item we have */
                switch(vmgram_tok_type(tokp))
                {
                case VMGRAM_MATCH_PROD:
                    /* 
                     *   We can't rule out a production without checking
                     *   it fully, which we're not doing on this pass;
                     *   however, note that we found a production, because
                     *   it means that we must from now on try skipping
                     *   any number of tokens before ruling out a match on
                     *   other grounds, since the production could
                     *   potentially match any number of tokens.
                     *   
                     *   If we're doing circular matches only, and this is
                     *   a circular production, and we're on the first
                     *   token in our list, we already know it matches, so
                     *   don't even count it as a production before the
                     *   item.  
                     */
                    if (is_circ && circ_only && cur_alt_tok == 0)
                    {
                        /* 
                         *   ignore the fact that it's a production, since
                         *   we already know it matches - we don't want to
                         *   be flexible about tokens following this item
                         *   since we know the exact number that match it 
                         */
                    }
                    else
                    {
                        /* note the production item */
                        prod_before = TRUE;
                    }

                    /* 
                     *   we can ignore this item for the purposes of
                     *   detecting a mismatch, because we can't tell
                     *   during this superficial scan whether it matches 
                     */
                    ignore_item = TRUE;
                    break;
                
                case VMGRAM_MATCH_SPEECH:
                    /* we must match a part of speech */
                    if (find_prop_in_tok(&tok[tok_idx],
                                         vmgram_tok_voc_prop(tokp)))
                    {
                        /* it's a match */
                        match = TRUE;
                    }
                    break;
                    
                case VMGRAM_MATCH_LITERAL:
                    /* we must match a literal */
                    if (tok[tok_idx].txt_ != 0
                        && tok_equals_lit(tok[tok_idx].txt_,
                                          tok[tok_idx].len_,
                                          vmgram_tok_lit_txt(tokp),
                                          vmgram_tok_lit_len(tokp),
                                          trunc_len))
                    {
                        /* it's a match */
                        match = TRUE;
                    }
                    break;
                
                case VMGRAM_MATCH_TOKTYPE:
                    /* we must match a token type */
                    if (tok[tok_idx].typ_ == vmgram_tok_tok_enum(tokp))
                        match = TRUE;
                    break;

                case VMGRAM_MATCH_STAR:
                    /* consume everything remaining on the line */
                    tok_idx = tok_cnt;

                    /* this is a match */
                    match = TRUE;
                    match_star = TRUE;
                    break;
                }

                /* if we found a match, we can stop scanning now */
                if (match)
                    break;

                /* 
                 *   we didn't match this token - if we have a production
                 *   before this point, we must keep scanning token, since
                 *   the production could end up matching any number of
                 *   tokens; if we don't have a production item, though,
                 *   we don't need to look any further, and we can rule
                 *   out this alternative immediately 
                 */
                if (!prod_before)
                    break;

                /* 
                 *   if we're ignoring this item (because we can't decide
                 *   now whether it's a match or not, and hence can't rule
                 *   it out), don't scan any tokens for the item 
                 */
                if (ignore_item)
                    break;
            }

            /* check to see if we found a match */
            if (match)
            {
                /* 
                 *   we found a match - skip the current token (unless we
                 *   matched a '*' item, in which case we've already
                 *   consumed all remaining tokens) 
                 */
                if (!match_star)
                    ++tok_idx;
            }
            else if (!ignore_item)
            {
                /* 
                 *   we didn't match, and we can't ignore this item - note
                 *   that the alternative does not match 
                 */
                found_mismatch = TRUE;
            }
        }

        /* 
         *   If we didn't find a reason to rule out this alternative,
         *   enqueue the alternative.  Never enqueue circular references,
         *   unless we're ONLY enqueuing circular references.
         */
        if (!found_mismatch
            && ((!circ_only && !is_circ)
                || (circ_only && is_circ)))
        {
            CVmGramProdState *state;
            
            /* create and enqueue the new state */
            state = enqueue_new_state(mem, start_tok_pos, enclosing_state,
                                      altp, self, &need_to_clone, queues,
                                      found_circ);

            /* 
             *   if we are enqueuing circular references, we already have
             *   a match for the first (circular) token, so fill it in 
             */
            if (circ_only)
            {
                CVmGramProdMatch *match;
                
                /* create a match for the alternative */
                match = CVmGramProdMatch::
                        alloc(mem, 0, vmgram_tok_prop(first_tokp),
                              circ_match->proc_obj_,
                              circ_match->sub_match_list_,
                              circ_match->sub_match_cnt_);

                /* set the first match in the token */
                state->match_list_[0] = match;

                /* set up at the second token */
                state->alt_pos_ = 1;
            }
        }

        /* 
         *   move on to the next alternative - it's immediately after the
         *   last token of the current alternative 
         */
        altp = tokp;
    }
}

/*
 *   Create and enqueue a new state 
 */
CVmGramProdState *CVmObjGramProd::
   enqueue_new_state(CVmGramProdMem *mem,
                     size_t start_tok_pos,
                     CVmGramProdState *enclosing_state,
                     const char *altp, vm_obj_id_t self,
                     int *need_to_clone, CVmGramProdQueue *queues,
                     int circular_alt)
{
    CVmGramProdState *state;

    /* create the new state */
    state = create_new_state(mem, start_tok_pos, enclosing_state,
                             altp, self, need_to_clone, circular_alt);
    
    /* 
     *   Add the item to the appropriate queue.  If the item has an
     *   associated badness, add it to the badness queue.  Otherwise, add
     *   it to the work queue.  
     */
    if (vmgram_alt_badness(altp) != 0)
    {
        /* 
         *   we have a badness rating - add it to the badness queue, since
         *   we don't want to process it until we entirely better
         *   possibilities 
         */
        state->nxt_ = queues->badness_queue_;
        queues->badness_queue_ = state;
    }
    else
    {
        /* enqueue the state */
        enqueue_state(state, queues);
    }

    /* return the new state */
    return state;
}

/*
 *   Create a new state 
 */
CVmGramProdState *CVmObjGramProd::
   create_new_state(CVmGramProdMem *mem, size_t start_tok_pos,
                    CVmGramProdState *enclosing_state,
                    const char *altp, vm_obj_id_t self,
                    int *need_to_clone, int circular_alt)
{
    /* 
     *   if necessary, clone the enclosing state; we need to do this if we
     *   enqueue more than one nested alternative, since each nested
     *   alternative could parse the rest of the token list differently
     *   and thus provide a different result to the enclosing state 
     */
    if (*need_to_clone && enclosing_state != 0)
        enclosing_state = enclosing_state->clone(mem);

    /* 
     *   we will need to clone the enclosing state if we create another
     *   alternative after this one 
     */
    *need_to_clone = TRUE;

    /* create a new state object for the alternative */
    return CVmGramProdState::alloc(mem, start_tok_pos, 0,
                                   enclosing_state, altp, self,
                                   circular_alt);
}

/*
 *   Enqueue a state in the work queue
 */
void CVmObjGramProd::enqueue_state(CVmGramProdState *state,
                                   CVmGramProdQueue *queues)
{
    /* add it to the work queue */
    state->nxt_ = queues->work_queue_;
    queues->work_queue_ = state;
}

/*
 *   Determine if a token from the input matches a literal token, allowing
 *   the input token to be a truncated version of the literal token if the
 *   given truncation length is non-zero. 
 */
int CVmObjGramProd::tok_equals_lit(const char *tok, size_t tok_len,
                                   const char *lit, size_t lit_len,
                                   size_t trunc_len)
{
    size_t compare_len;

    /* 
     *   If truncation is allowed, and if the input token is at least as
     *   long as the truncation length, and the literal is at least as
     *   long as the token, and the token is the leading substring of the
     *   literal, they match; otherwise, they match only if they match
     *   exactly.  
     */
    if (trunc_len != 0 && tok_len >= trunc_len && lit_len > tok_len)
    {
        /* 
         *   truncation is allowed, and the input token is longer than the
         *   minimum length for truncation, and the literal is even longer
         *   - compare only up to the length of the input token 
         */
        compare_len = tok_len;
    }
    else
    {
        /* 
         *   we can't truncate the token - compare it to the entire
         *   literal 
         */
        compare_len = lit_len;
    }
    
    /* check for a match */
    return (tok_len == compare_len
            && memicmp(tok, lit, compare_len) == 0);
}


/*
 *   Find a property in a token's list.  Returns true if the property is
 *   there, false if not. 
 */
int CVmObjGramProd::find_prop_in_tok(const vmgramprod_tok *tok,
                                     vm_prop_id_t prop)
{
    size_t i;
    const vm_prop_id_t *p;

    /* search for the property */
    for (i = tok->prop_cnt_, p = tok->props_ ; i != 0 ; --i, ++p)
    {
        /* if this is the one we're looking for, return success */
        if (*p == prop)
            return TRUE;
    }

    /* we didn't find it */
    return FALSE;
}

/*
 *   Get the next token item after the given item in an alternative
 */
const char *CVmObjGramProd::get_next_alt_tok(const char *tokp)
{
    switch(vmgram_tok_type(tokp))
    {
    case VMGRAM_MATCH_PROD:
        return tokp + VMGRAM_TOK_PROD_SIZE;
        
    case VMGRAM_MATCH_SPEECH:
        return tokp + VMGRAM_TOK_SPEECH_SIZE;

    case VMGRAM_MATCH_LITERAL:
        return tokp + VMGRAM_TOK_LIT_SIZE(tokp);

    case VMGRAM_MATCH_TOKTYPE:
        return tokp + VMGRAM_TOK_TYPE_SIZE;

    case VMGRAM_MATCH_STAR:
        return tokp + VMGRAM_TOK_STAR_SIZE;

    default:
        err_throw(VMERR_INVAL_METACLASS_DATA);
        return 0;
    }
}

/*
 *   Get the next alternative 
 */
const char *CVmObjGramProd::get_next_alt(const char *altp)
{
    size_t tok_cnt;
    const char *tokp;

    /* get the token count */
    tok_cnt = vmgram_alt_tokcnt(altp);

    /* skip the tokens */
    for (tokp = vmgram_alt_tokptr(altp) ; tok_cnt != 0 ;
         --tok_cnt, tokp = get_next_alt_tok(tokp)) ;

    /* the next alternative starts immediately after the last token */
    return tokp;
}

