/* $Header$ */

/* Copyright (c) 1999 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmmain.h - main entrypoint to run a T3 image file
Function
  
Notes
  
Modified
  10/07/99 MJRoberts  - Creation
*/

#ifndef VMMAIN_H
#define VMMAIN_H

#include "vmglob.h"

/*
 *   Execute an image file.  We'll return zero on success, or a VM error
 *   code on failure.  If an error occurs, we'll fill in 'errbuf' with the
 *   text of a message describing the problem.
 *   
 *   If 'load_from_exe' is true, the image filename given is actually the
 *   name of the native executable file that we're running, and we should
 *   load the image file that's attached to the native executable file via
 *   the system-specific os_exeseek() mechanism.
 *   
 *   If 'script_file' is not null, we'll read console input from the given
 *   file.  If 'log_file' is not null, we'll log console output to the
 *   given file.  If 'cmd_log_file' is not null, we'll log each line we
 *   read from the console to the given command logging file.  
 */
int vm_run_image(class CVmMainClientIfc *clientifc,
                 const char *image_file_name,
                 class CVmHostIfc *hostifc,
                 const char *const *prog_argv, int prog_argc,
                 const char *script_file, const char *log_file,
                 const char *cmd_log_file,
                 int load_from_exe, int show_banner, const char *charset,
                 const char *saved_state);

/*
 *   Execute an image file using argc/argv conventions.  We'll parse the
 *   command line and invoke the program.
 *   
 *   The 'executable_name' is the name of the host program; this is used
 *   to prepare "usage" messages.
 *   
 *   If 'defext' is true, we'll try adding a default extension ("t3x") to
 *   the name of the image file we find if the given filename doesn't
 *   exist.  We'll always check to see if the file exists with the exact
 *   given name before we do this, so that we don't add an extension where
 *   none is needed.  If the caller doesn't want us to try adding an
 *   extension at all, pass in 'defext' as false.
 *   
 *   If 'hostifc' is null, we'll provide our own default interface.  The
 *   caller can provide a custom host interface by passing in a non-null
 *   'hostifc' value.  
 */
int vm_run_image_main(class CVmMainClientIfc *clientifc,
                      const char *executable_name,
                      int argc, char **argv, int defext,
                      class CVmHostIfc *hostifc);

/*
 *   VM Main client services interface.  Callers of the vm_run_image
 *   functions must provide an implementation of this interface. 
 */
class CVmMainClientIfc
{
public:
    /* 
     *   initialize - we'll invoke this function just before starting
     *   execution in the loaded image
     */
    virtual void pre_exec(struct vm_globals *globals) = 0;

    /* 
     *   terminate - we'll invoke this just after execution in the loaded
     *   image terminates 
     */
    virtual void post_exec(struct vm_globals *globals) = 0;

    /* 
     *   Terminate with error - we'll invoke this upon catching an
     *   exception that the image file doesn't handle and which thus
     *   terminates execution.  Note that if this is called, post_exec()
     *   will not be called; however, if post_exec() itself throws an
     *   exception, we'll invoke this routine.  
     */
    virtual void post_exec_err(struct vm_globals *globals) = 0;

    /* 
     *   Display an error message.  We'll call this with a complete error
     *   message to display.  Note that we won't add a newline at the end
     *   of the message, so if the message is to be displayed on a
     *   stdio-style terminal, this routine should display a newline after
     *   the message.  
     */
    virtual void display_error(struct vm_globals *globals,
                               const char *msg) = 0;
};

#endif /* VMMAIN_H */

