#ifdef RCSID
static char RCSid[] =
"$Header$";
#endif

/* Copyright (c) 2000 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmvec.cpp - T3 VM Vector metaclass
Function
  
Notes
  
Modified
  05/14/00 MJRoberts  - Creation
*/

#include <stdlib.h>

#include "t3std.h"
#include "vmtype.h"
#include "vmmcreg.h"
#include "vmmeta.h"
#include "vmglob.h"
#include "vmobj.h"
#include "vmarray.h"
#include "vmvec.h"
#include "vmlst.h"
#include "vmerr.h"
#include "vmerrnum.h"
#include "vmfile.h"
#include "vmstack.h"
#include "vmbif.h"
#include "vmundo.h"
#include "vmrun.h"
#include "vmiter.h"
#include "vmsort.h"


/* ------------------------------------------------------------------------ */
/*
 *   statics 
 */

/* metaclass registration object */
static CVmMetaclassVector metaclass_reg_obj;
CVmMetaclass *CVmObjVector::metaclass_reg_ = &metaclass_reg_obj;

/* function table */
int (CVmObjVector::
     *CVmObjVector::func_table_[])(VMG_ vm_obj_id_t self,
                                   vm_val_t *retval, uint *argc) =
{
    &CVmObjVector::getp_undef,
    &CVmObjVector::getp_set_length,
    &CVmObjVector::getp_insert_at,
    &CVmObjVector::getp_remove_element_at,
    &CVmObjVector::getp_remove_range,
    &CVmObjVector::getp_append
};

/* ------------------------------------------------------------------------ */
/* 
 *   create with a given number of elements 
 */
CVmObjVector::CVmObjVector(VMG_ size_t element_count)
    : CVmObjArray(vmg_ element_count)
{
    /* 
     *   a vector intially has no elements in use; the element_count
     *   merely gives the number of slots to be allocated initially (this
     *   won't affect the allocation count, which is stored separately) 
     */
    set_element_count(0);
}

/* ------------------------------------------------------------------------ */
/* 
 *   create dynamically using stack arguments 
 */
vm_obj_id_t CVmObjVector::create_from_stack(VMG_ const uchar **pc_ptr,
                                            uint argc)
{
    vm_val_t *arg1;
    vm_obj_id_t id;
    CVmObjVector *vec;
    
    /* check arguments */
    if (argc != 1)
        err_throw(VMERR_WRONG_NUM_OF_ARGS);

    /* get the first argument */
    arg1 = G_stk->get(0);

    /* check what kind of argument we have */
    if (arg1->typ == VM_INT)
    {
        size_t cnt;
        
        /* get the number of elements to allocate */
        cnt = (size_t)arg1->val.intval;

        /* create the vector with the given number of elements */
        id = vm_new_id(vmg_ FALSE);
        vec = new (vmg_ id) CVmObjVector(vmg_ cnt);
    }
    else
    {
        /* anything else is invalid */
        err_throw(VMERR_BAD_TYPE_BIF);
    }

    /* discard arguments */
    G_stk->discard(argc);

    /* return the new object */
    return id;
}

/* ------------------------------------------------------------------------ */
/* 
 *   load from an image file 
 */
void CVmObjVector::load_from_image(VMG_ vm_obj_id_t self,
                                   const char *ptr, size_t siz)
{
    size_t alloc_cnt;
    size_t ele_cnt;

    /* if we already have memory allocated, free it */
    if (ext_ != 0)
    {
        G_mem->get_var_heap()->free_mem(ext_);
        ext_ = 0;
    }

    /* get the allocation size and the actual in-use size */
    alloc_cnt = vmb_get_len(ptr);
    ele_cnt = vmb_get_len(ptr + VMB_LEN);

    /* make sure we allocate at least as much as is actually in use */
    if (alloc_cnt < ele_cnt)
        alloc_cnt = ele_cnt;

    /* make sure the size isn't larger than we'd expect */
    if (siz > VMB_LEN + (VMB_DATAHOLDER * ele_cnt))
        siz = VMB_LEN + (VMB_DATAHOLDER * ele_cnt);

    /* 
     *   allocate memory at the new allocation size as indicated in the
     *   image data 
     */
    alloc_array(vmg_ alloc_cnt);

    /* set the in-use element count */
    set_element_count(ele_cnt);

    /* 
     *   if the size is smaller than we'd expect from the initialized
     *   element count, set extra elements to nil 
     */
    if (siz < VMB_LEN + (VMB_DATAHOLDER * ele_cnt))
    {
        size_t i;
        vm_val_t nil_val;

        /* set all elements to nil */
        nil_val.set_nil();
        for (i = 0 ; i < ele_cnt ; ++i)
            set_element(i, &nil_val);
    }

    /* copy the initialized elements from the image data */
    memcpy(get_element_ptr(0), ptr + VMB_LEN, siz - VMB_LEN);
}

/* ------------------------------------------------------------------------ */
/* 
 *   save to a file 
 */
void CVmObjVector::save_to_file(VMG_ class CVmFile *fp)
{
    size_t ele_cnt;
    size_t alloc_cnt;

    /* get our element count and full allocation count */
    ele_cnt = get_element_count();
    alloc_cnt = get_allocated_count();

    /* write the counts and the elements */
    fp->write_int2(alloc_cnt);
    fp->write_int2(ele_cnt);
    fp->write_bytes(get_element_ptr(0), calc_alloc_ele(ele_cnt));
}

/* 
 *   restore from a file 
 */
void CVmObjVector::restore_from_file(VMG_ vm_obj_id_t self,
                                     class CVmFile *fp, CVmObject *obj)
{
    size_t ele_cnt;
    size_t alloc_cnt;

    /* read the element count and the full allocation count */
    ele_cnt = fp->read_uint2();
    alloc_cnt = fp->read_uint2();

    /* free any existing extension */
    if (ext_ != 0)
    {
        G_mem->get_var_heap()->free_mem(ext_);
        ext_ = 0;
    }

    /* allocate the space at the allocation count */
    alloc_array(vmg_ alloc_cnt);

    /* set the actual in-use element count */
    set_element_count(ele_cnt);

    /* read the contents */
    fp->read_bytes(get_element_ptr(0), calc_alloc_ele(ele_cnt));
}

/* ------------------------------------------------------------------------ */
/* 
 *   create with no initial contents 
 */
vm_obj_id_t CVmObjVector::create(VMG_ int in_root_set)
{
    vm_obj_id_t id = vm_new_id(vmg_ in_root_set);
    new (vmg_ id) CVmObjVector();
    return id;
}

/* 
 *   create with a given number of elements 
 */
vm_obj_id_t CVmObjVector::create(VMG_ int in_root_set, size_t element_count)
{
    vm_obj_id_t id = vm_new_id(vmg_ in_root_set);
    new (vmg_ id) CVmObjVector(vmg_ element_count);
    return id;
}

/* ------------------------------------------------------------------------ */
/* 
 *   get a property 
 */
int CVmObjVector::get_prop(VMG_ vm_prop_id_t prop, vm_val_t *retval,
                           vm_obj_id_t self, vm_obj_id_t *source_obj,
                           uint *argc)
{
    ushort func_idx;

    /* if we succeed, the result will come directly from this object */
    *source_obj = self;

    /* translate the property index to an index into our function table */
    func_idx = G_meta_table
               ->prop_to_vector_idx(metaclass_reg_->get_reg_idx(), prop);

    /* call the appropriate function */
    if ((this->*func_table_[func_idx])(vmg_ self, retval, argc))
        return TRUE;

    /* inherit array handling */
    return CVmObjArray::get_prop(vmg_ prop, retval, self, source_obj, argc);
}

/* ------------------------------------------------------------------------ */
/*
 *   Expand the vector to accomodate the given number of new elements.
 *   We'll increase the element count, and we'll save undo for the change.
 *   If necessary, we'll re-allocate our memory extension block at a
 *   larger size. 
 */
void CVmObjVector::expand_by(VMG_ vm_obj_id_t self, size_t added_elements)
{
    size_t new_ele_cnt;

    /* calculate the element count */
    new_ele_cnt = get_element_count() + added_elements;
    
    /* 
     *   if the new size is larger than our current allocation size,
     *   increase the allocated size 
     */
    if (new_ele_cnt > get_allocated_count())
    {
        size_t new_alloc_cnt;
        size_t new_mem_size;

        /* bump up the allocated size */
        new_alloc_cnt = get_allocated_count() + get_alloc_count_increment();
        
        /* get the new size we need */
        new_mem_size = calc_alloc(new_alloc_cnt);

        /* reallocate our memory at the new, larger size */
        ext_ = (char *)G_mem->get_var_heap()
               ->realloc_mem(new_mem_size, ext_, this);

        /* set the new allocation size */
        set_allocated_count(new_alloc_cnt);
    }

    /* remember the new size, saving undo */
    set_element_count_undo(vmg_ self, new_ele_cnt);
}


/* ------------------------------------------------------------------------ */
/*
 *   Set the element count, keeping undo for the change 
 */
void CVmObjVector::set_element_count_undo(VMG_ vm_obj_id_t self,
                                          size_t new_ele_cnt)
{
    /* add undo if necessary */
    if (G_undo != 0)
    {
        vm_val_t old_val;
        
        /* set up an integer with the pre-modification element count */
        old_val.set_int(get_element_count());

        /* 
         *   add the undo record - use the special index value 0xffffffff
         *   as the key; this will never be a valid index for a real
         *   element, because we could never address an element with an
         *   index this high in a 32-bit address space 
         */
        G_undo->add_new_record_int_key(vmg_ self, 0xffffffffU, &old_val);
    }

    /* set the new size */
    set_element_count(new_ele_cnt);
}

/* ------------------------------------------------------------------------ */
/* 
 *   add a value to the vector 
 */
void CVmObjVector::add_val(VMG_ vm_val_t *result,
                           vm_obj_id_t self, const vm_val_t *val)
{
    const char *lst;
    size_t idx;

    /* the result is always myself, suitably modified */
    result->set_obj(self);

    /* 
     *   remember the index of the first new element - this is simply one
     *   higher than the last valid current index 
     */
    idx = get_element_count();

    /* check to see if the value is a list */
    if ((lst = val->get_as_list(vmg0_)) != 0)
    {
        size_t i;
        
        /* expand the vector by the number of elements in the list */
        expand_by(vmg_ self, vmb_get_len(lst));

        /* add each element of the list */
        for (i = 0 ; i < vmb_get_len(lst) ; ++i)
        {
            vm_val_t ele_val;

            /* get this list element */
            CVmObjList::index_list(vmg_ &ele_val, lst, i + 1);

            /* store this element in the vector */
            set_element_undo(vmg_ self, idx + i, &ele_val);
        }
    }
    else
    {
        /* expand the vector by one element */
        expand_by(vmg_ self, 1);

        /* 
         *   set the newly-created last element of the vector to the given
         *   value 
         */
        set_element_undo(vmg_ self, idx, val);
    }
}

/* ------------------------------------------------------------------------ */
/* 
 *   subtract a value from the vector 
 */
void CVmObjVector::sub_val(VMG_ vm_val_t *result,
                           vm_obj_id_t self, const vm_val_t *val)
{
    const char *lst;
    size_t ele_cnt;

    /* the result is always myself, suitably modified */
    result->set_obj(self);

    /* note the initial element count */
    ele_cnt = get_element_count();

    /* check to see if the value is a list */
    if ((lst = val->get_as_list(vmg0_)) != 0)
    {
        size_t i;

        /* subtract each element of the list */
        for (i = 0 ; i < vmb_get_len(lst) ; ++i)
        {
            vm_val_t ele_val;

            /* 
             *   get the list pointer again, in case we had to swap
             *   constant pool pages 
             */
            lst = val->get_as_list(vmg0_);

            /* get this list element */
            CVmObjList::index_list(vmg_ &ele_val, lst, i + 1);

            /* subtract it */
            ele_cnt = sub_one_val(vmg_ self, &ele_val, ele_cnt);
        }
    }
    else
    {
        /* subtract the single value */
        ele_cnt = sub_one_val(vmg_ self, val, ele_cnt);
    }

    /* update the stored element count */
    set_element_count_undo(vmg_ self, ele_cnt);
}

/*
 *   Subtract an element from the values.  This doesn't use or change the
 *   stored element count, but rather uses the given element count and
 *   returns the new element count; this allows this routine to be called
 *   repeatedly to subtract a series of values from the vector without
 *   creating individual undo records for each subtraction.  The caller is
 *   responsible for storing the final element count when done with the
 *   entire bundle of subtractions.  
 */
size_t CVmObjVector::sub_one_val(VMG_ vm_obj_id_t self,
                                 const vm_val_t *val, size_t ele_cnt)
{
    size_t src;
    size_t dst;
    size_t idx;
    vm_val_t nil_val;
    
    /* search the array for the given value, and remove each occurrence */
    for (src = dst = 0 ; src < ele_cnt ; ++src)
    {
        vm_val_t ele;
        
        /* get this element */
        get_element(src, &ele);
        
        /* check for a match to the value to be subtracted */
        if (!ele.equals(vmg_ val))
        {
            /* 
             *   this element must be copied to the output - if the
             *   destination and source pointers aren't the same, copy
             *   this source value to its new position (note that we
             *   specifically do not perform the copy if it the pointers
             *   are the same so that we can avoid creating unnecessary
             *   undo records) 
             */
            if (dst != src)
                set_element_undo(vmg_ self, dst, &ele);

            /* adjust the destination index for the added value */
            ++dst;
        }
    }

    /*
     *   Fill out the rest of the elements (the ones that will be
     *   eliminated after we set the new vector size) with nil.  This will
     *   ensure that we create an undo record for each of the element
     *   slots we're dropping, so that they'll be restored to their
     *   pre-shrinkage values if we undo. 
     */
    nil_val.set_nil();
    for (idx = dst ; idx < ele_cnt ; ++idx)
        set_element_undo(vmg_ self, idx, &nil_val);

    /* return the number of elements we retained */
    return dst;
}

/* ------------------------------------------------------------------------ */
/*
 *   apply undo 
 */
void CVmObjVector::apply_undo(VMG_ struct CVmUndoRecord *rec)
{
    /* 
     *   if the record refers to index 0xffffffff, this is a size-change
     *   record; otherwise, it's an ordinary element change record 
     */
    if ((size_t)rec->id.intval == 0xffffffffU)
    {
        /* 
         *   it's a size change - apply the new size, which is stored as
         *   an integer in the old value record 
         */
        set_element_count((size_t)rec->oldval.val.intval);
    }
    else
    {
        /* 
         *   it's an ordinary index record - inherit the default Array
         *   handling 
         */
        CVmObjArray::apply_undo(vmg_ rec);
    }
}

/* ------------------------------------------------------------------------ */
/* 
 *   property evaluator - set the length 
 */
int CVmObjVector::getp_set_length(VMG_ vm_obj_id_t self, vm_val_t *retval,
                                  uint *argc)
{
    size_t old_len;
    size_t new_len;
    size_t idx;
    vm_val_t nil_val;
    
    /* check arguments */
    if (get_prop_check_argc(retval, argc, 1))
        return TRUE;

    /* the return value is 'self' */
    retval->set_obj(self);

    /* note the old length */
    old_len = get_element_count();

    /* get the new length */
    new_len = (size_t)CVmBif::pop_int_val(vmg0_);

    /* set up the nil value for filling new/dropped elements */
    nil_val.set_nil();

    /* determine if we're growing or shrinking the value */
    if (new_len > old_len)
    {
        /* we're growing - increase the vector to its new size */
        set_element_count_undo(vmg_ self, new_len);

        /* 
         *   Set each newly added element to nil.  Note that we don't
         *   bother saving undo for these changes: to undo this change,
         *   the only thing we'll have to do is reduce the vector size to
         *   its previous size, and we've already saved undo for the size
         *   change. 
         */
        for (idx = old_len ; idx < new_len ; ++idx)
            set_element(idx, &nil_val);
    }
    else
    {
        /* 
         *   We're shrinking the vector.  For each element that we're
         *   losing, save undo for setting the element; to do this, simply
         *   set each element we're losing to nil, since this will add an
         *   undo record with the original value of the element.  
         */
        for (idx = new_len ; idx < old_len ; ++idx)
            set_element_undo(vmg_ self, idx, &nil_val);

        /*
         *   Set the new size of the vector.  Note that we do this AFTER
         *   we have saved undo for the elements.  We must perform the
         *   steps in this order because undo is applied in reverse order:
         *   when we read back the undo, we'll first change the vector
         *   size to its old size (since that's the last saved undo
         *   instruction), then we'll set the elements to their old
         *   values. 
         */
        set_element_count_undo(vmg_ self, new_len);
    }

    /* handled */
    return TRUE;
}

/* ------------------------------------------------------------------------ */
/* 
 *   property evaluator - insert one or more elements at a given index
 */
int CVmObjVector::getp_insert_at(VMG_ vm_obj_id_t self, vm_val_t *retval,
                                 uint *in_argc)
{
    size_t idx;
    size_t ins_idx;
    size_t old_cnt;
    size_t add_cnt;
    size_t new_cnt;
    uint argc = (in_argc != 0 ? *in_argc : 0);

    /* we must have at least two arguments */
    if (get_prop_check_argc_range(retval, in_argc, 2, 32767))
        return TRUE;

    /* get the starting insertion index */
    ins_idx = (size_t)CVmBif::pop_int_val(vmg0_);

    /* the return value is 'self' */
    retval->set_obj(self);

    /* note the original size of the array */
    old_cnt = get_element_count();

    /* 
     *   note the number of items we're adding - we're adding each
     *   argument, other than the first 
     */
    add_cnt = argc - 1;

    /* 
     *   note the new size - we're inserting one element for each argument
     *   past the first one 
     */
    new_cnt = old_cnt + add_cnt;

    /* 
     *   if the starting index is out of range, it's an error - it can
     *   range from 1 to one higher than the current element count (the
     *   top end of the range allows us to insert elements after all of
     *   the current elements) 
     */
    if (ins_idx < 1 || ins_idx > get_element_count() + 1)
        err_throw(VMERR_INDEX_OUT_OF_RANGE);

    /* adjust the starting index to a zero-based value */
    --ins_idx;

    /* create the new elements */
    set_element_count_undo(vmg_ self, new_cnt);

    /* 
     *   Move the existing elements to their new slots.  Start with the
     *   starting index, and move it to its new position, since the
     *   starting index is the first element that needs to be moved.  Note
     *   that we start at the top and work down, to avoid overwriting any
     *   overlapping part of the new position block and the old position
     *   block.  
     */
    idx = old_cnt;
    for (;;)
    {
        size_t new_idx;
        vm_val_t ele;

        /* move to the previous element */
        --idx;

        /* get the value at this slot */
        get_element(idx, &ele);
        
        /* calculate the new slot */
        new_idx = idx + add_cnt;

        /* 
         *   if the new index is within the old size range, keep undo for
         *   the change; otherwise, no undo is necessary, since when we
         *   undo we'll be completely dropping this new slot 
         */
        if (new_idx < old_cnt)
            set_element_undo(vmg_ self, new_idx, &ele);
        else
            set_element(new_idx, &ele);

        /* if this was the last element, we're done */
        if (idx == ins_idx)
            break;
    }

    /* add the new items */
    for (idx = ins_idx ; add_cnt != 0 ; ++idx, --add_cnt)
    {
        vm_val_t val;
        
        /* pop the next argument value */
        G_stk->pop(&val);

        /* store the item, keeping undo only if it's in the old size range */
        if (idx < old_cnt)
            set_element_undo(vmg_ self, idx, &val);
        else
            set_element(idx, &val);
    }

    /* handled */
    return TRUE;
}

/* ------------------------------------------------------------------------ */
/* 
 *   property evaluator - append an element
 */
int CVmObjVector::getp_append(VMG_ vm_obj_id_t self, vm_val_t *retval,
                              uint *argc)
{
    vm_val_t val;
    size_t cnt;

    /* check arguments */
    if (get_prop_check_argc(retval, argc, 1))
        return TRUE;

    /* get the old size */
    cnt = get_element_count();

    /* expand the vector by one element */
    set_element_count_undo(vmg_ self, cnt + 1);

    /* retrieve the value for the new element */
    G_stk->pop(&val);

    /* 
     *   set the last element - note that there's no need to save undo,
     *   since undoing this operation will simply discard this element (in
     *   other words, there's no previous value for this element since
     *   it's being created anew here) 
     */
    set_element(cnt, &val);

    /* the return value is 'self' */
    retval->set_obj(self);

    /* handled */
    return TRUE;
}

/* ------------------------------------------------------------------------ */
/* 
 *   property evaluator - remove an element at the given position
 */
int CVmObjVector::getp_remove_element_at(VMG_ vm_obj_id_t self,
                                         vm_val_t *retval, uint *argc)
{
    size_t start_idx;

    /* check arguments */
    if (get_prop_check_argc(retval, argc, 1))
        return TRUE;

    /* get the deletion index */
    start_idx = (size_t)CVmBif::pop_int_val(vmg0_);

    /* make sure the index is in range */
    if (start_idx < 1 || start_idx > get_element_count())
        err_throw(VMERR_INDEX_OUT_OF_RANGE);

    /* adjust to a zero-based index */
    --start_idx;

    /* the return value is 'self' */
    retval->set_obj(self);

    /* delete one element */
    remove_elements_undo(vmg_ self, start_idx, 1);


    /* handled */
    return TRUE;
}

/* ------------------------------------------------------------------------ */
/*
 *   property evaluator - remove a range of elements
 */
int CVmObjVector::getp_remove_range(VMG_ vm_obj_id_t self,
                                    vm_val_t *retval, uint *argc)
{
    size_t start_idx;
    size_t end_idx;

    /* check arguments */
    if (get_prop_check_argc(retval, argc, 2))
        return TRUE;

    /* get the starting index */
    start_idx = (size_t)CVmBif::pop_int_val(vmg0_);

    /* 
     *   make sure the index is in range - it must refer to an existing
     *   element 
     */
    if (start_idx < 1 || start_idx > get_element_count())
        err_throw(VMERR_INDEX_OUT_OF_RANGE);

    /* get the ending index */
    end_idx = (size_t)CVmBif::pop_int_val(vmg0_);

    /* 
     *   make sure it's in range - it must refer to an existing element,
     *   and it must be greater than or equal to the starting index 
     */
    if (end_idx < start_idx || end_idx > get_element_count())
        err_throw(VMERR_INDEX_OUT_OF_RANGE);

    /* adjust to zero-based indices */
    --start_idx;
    --end_idx;

    /* the return value is 'self' */
    retval->set_obj(self);

    /* 
     *   delete the elements - the number of elements we're deleting is
     *   one higher than the difference of the starting and ending indices
     *   (if the two indices are the same, we're deleting just the one
     *   element) 
     */
    remove_elements_undo(vmg_ self, start_idx, end_idx - start_idx + 1);

    /* handled */
    return TRUE;
}

/* ------------------------------------------------------------------------ */
/*
 *   Delete a range of elements 
 */
void CVmObjVector::remove_elements_undo(VMG_ vm_obj_id_t self,
                                        size_t start_idx, size_t del_cnt)
{
    size_t idx;
    size_t cnt;
    
    /* note the original size of the array */
    cnt = get_element_count();

    /* move all of the elements to their new slots, keeping undo */
    for (idx = start_idx ; idx < cnt ; ++idx)
    {
        vm_val_t val;
        size_t move_idx;
        
        /* 
         *   calculate the index of the element to replace this one - it's
         *   past us by the number of items being deleted 
         */
        move_idx = idx + del_cnt;

        /* 
         *   if we're moving from a valid index in the old vector, get the
         *   element; otherwise, replace the value with nil (but still
         *   replace the value, since we want to keep undo for its
         *   original value) 
         */
        if (move_idx < cnt)
            get_element(move_idx, &val);
        else
            val.set_nil();

        /* store the element at its new position */
        set_element_undo(vmg_ self, idx, &val);
    }

    /*
     *   Reduce the size, keeping undo.  Note that it's important we do
     *   this last: undo is applied in reverse order, so when applying
     *   undo, we first want to increase the vector's size to its original
     *   size, then we want to apply the element value restorations.
     */
    set_element_count_undo(vmg_ self, cnt - del_cnt);
}

